#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the "theanolm decode" command.
"""

import logging

from theanolm.backend import TextFileType
from theanolm.scoring.slflattice import SLFLattice
from theanolm.scoring.kaldilattice import KaldiLattice, read_kaldi_vocabulary

class LatticeBatch(object):
    def __init__(self, lattices, lattice_list_file, lattice_format,
                 kaldi_vocabulary=None, num_jobs=1, job_id=0):
        """Reads the Kaldi word ID mapping, if given, and slices the lattices
        corresponding to the given job ID.

        If there's nothing else in the lattice list, adds the standard input.

        :type lattices: list of strs
        :param lattices: a list of lattice file paths

        :type lattice_list_file: file object
        :param lattice_list_file: a file containing paths to lattice files that
                                  will be added to the lattice list

        :type lattice_format: str
        :param lattice_format: format in which the lattices are saved; either
                               ``slf`` or ``kaldi``

        :type kaldi_vocabulary: file object
        :param kaldi_vocabulary: if not ``None``, the word to ID mapping for
                                 Kaldi lattices will be read from this file

        :type num_jobs: int
        :param num_jobs: split the lattice files to this many jobs

        :type job_id: int
        :param job_id: a number between ``0`` and ``num_jobs - 1``; select the
                       lattices corresponding to this job
        """

        # Read Kaldi word ID mapping.
        if kaldi_vocabulary is not None:
            self.kaldi_word_to_id = read_kaldi_vocabulary(kaldi_vocabulary)
            self.kaldi_id_to_word = [None] * len(self.kaldi_word_to_id)
            for word, id in self.kaldi_word_to_id.items():
                self.kaldi_id_to_word[id] = word
        elif lattice_format == 'kaldi':
            raise ValueError("Kaldi lattice vocabulary is not given.")

        if (lattice_format != 'slf') and (lattice_format != 'kaldi'):
            raise ValueError("Invalid lattice format specified ({})."
                             .format(lattice_format))
        self._lattice_format = lattice_format

        # Combine paths from command line and lattice list.
        if lattice_list_file is not None:
            lattices.extend(lattice_list_file.readlines())
        lattices = [path.strip() for path in lattices]
        # Ignore empty lines in the lattice list.
        lattices = [x for x in lattices if x]
        if not lattices:
            lattices = ["-"]

        # Pick every ith lattice, if --num-jobs is specified and > 1.
        if num_jobs < 1:
            raise ValueError("Invalid number of jobs specified ({})."
                             .format(num_jobs))
        if (job_id < 0) or (job_id > num_jobs - 1):
            raise ValueError("Invalid job selected ({})."
                             .format(job_id))
        self._lattices = lattices[job_id::num_jobs]

    def __iter__(self):
        """A generator for iterating through the lattices of this job.
        """

        file_type = TextFileType('r')

        for path in self._lattices:
            logging.info("Reading lattice file `%s´.", path)
            lattice_file = file_type(path)
            if self._lattice_format == 'slf':
                yield SLFLattice(lattice_file)
            else:
                assert self._lattice_format == 'kaldi'
                lattice_lines = []
                id_to_word = self.kaldi_id_to_word
                while True:
                    line = lattice_file.readline()
                    if not line:
                        # end of file
                        if lattice_lines:
                            yield KaldiLattice(lattice_lines, id_to_word)
                        break
                    line = line.strip()
                    if not line:
                        # empty line
                        if lattice_lines:
                            yield KaldiLattice(lattice_lines, id_to_word)
                        lattice_lines = []
                        continue
                    lattice_lines.append(line)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Modules related to text and word lattice scoring.
"""

from theanolm.scoring.textscorer import TextScorer
from theanolm.scoring.latticedecoder import LatticeDecoder
from theanolm.scoring.latticebatch import LatticeBatch
from theanolm.scoring.rescoredlattice import RescoredLattice



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Modules that implement a neural network and its layers.
"""

from theanolm.network.network import Network
from theanolm.network.architecture import Architecture
from theanolm.network.recurrentstate import RecurrentState



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the BidirectionalLayer class, a wrapper for creating
bidirectional layers.
"""

from copy import copy

import theano.tensor as tensor

from theanolm.network.grulayer import GRULayer
from theanolm.network.lstmlayer import LSTMLayer

class BidirectionalLayer(object):
    """Wrapper for Combining Forward and Backward Recurrent Layers

    M. Schuster, K. K. Paliwal
    Bidirectional Recurrent Neural Networks
    IEEE Transactions on Signal Processing, 45(11), 2673–2681

    Combines two recurrent layers, one which has dependency forward in time, and
    one with dependency backward in time. The input of the backward layer is
    shifted two time steps to make sure the target word is not predicted using
    itself (which is also the next input word). Note that the probability of a
    word depends on the future words as well, instead of just the past words.
    Thus the sequence probabilities are not a true probability distribution, and
    text cannot be generated.
    """

    def __init__(self, layer_options, *args, **kwargs):
        layer_type = layer_options['type']
        self.name = layer_options['name']
        if 'size' in layer_options:
            self.output_size = int(layer_options['size'])
        else:
            input_layers = layer_options['input_layers']
            self.output_size = sum([x.output_size for x in input_layers])
        backward_size = self.output_size // 2
        forward_size = self.output_size - backward_size

        forward_options = layer_options
        backward_options = copy(layer_options)
        forward_options['name'] = self.name + '/forward'
        forward_options['size'] = forward_size
        backward_options['name'] = self.name + '/backward'
        backward_options['size'] = backward_size
        backward_options['reverse_time'] = True
        if layer_type == 'blstm':
            self._forward_layer = LSTMLayer(forward_options, *args, **kwargs)
            self._backward_layer = LSTMLayer(backward_options, *args, **kwargs)
        elif layer_type == 'bgru':
            self._forward_layer = GRULayer(forward_options, *args, **kwargs)
            self._backward_layer = GRULayer(backward_options, *args, **kwargs)
        else:
            raise ValueError("Invalid layer type requested: " + layer_type)

        self.output = None

    def create_structure(self):
        """Creates the symbolic graph of this layer.

        Sets self.output to a symbolic matrix that describes the output of this
        layer.
        """

        self._forward_layer.create_structure()
        self._backward_layer.create_structure()
        self.output = tensor.concatenate([self._forward_layer.output,
                                          self._backward_layer.output],
                                         axis=2)

    def get_state(self, state):
        """Pulls parameter values from Theano shared variables.

        If there already is a parameter in the state, it will be replaced, so it
        has to have the same number of elements.

        :type state: h5py.File
        :param state: HDF5 file for storing the neural network parameters
        """

        self._forward_layer.get_state(state)
        self._backward_layer.get_state(state)

    def set_state(self, state):
        """Sets the values of Theano shared variables.

        :type state: h5py.File
        :param state: HDF5 file that contains the neural network parameters
        """

        self._forward_layer.set_state(state)
        self._backward_layer.set_state(state)

    def num_params(self):
        """Returns the number of parameters in this layer.

        This method is used just for reporting the number of parameters in the
        model. Normally there is just one set of parameters.

        :rtype: int
        :returns: the number of parameters used by the layer
        """

        return self._forward_layer.num_params() + \
               self._backward_layer.num_params()

    def get_variables(self):
        """Returns a dictionary of the shared variables.

        This function is used by the optimizers to create optimization
        parameters that are specific to network parameters, and compute
        gradients with regard to the parameters.

        :rtype: dict
        :returns: mapping from parameter path to Theano shared variables
        """

        result = dict()
        result.update(self._forward_layer.get_variables())
        result.update(self._backward_layer.get_variables())
        return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the "theanolm train" command.
"""

import sys
import mmap
import logging

import h5py
import numpy
import theano

from theanolm import Vocabulary, Architecture, Network
from theanolm.backend import TextFileType, get_default_device
from theanolm.parsing import LinearBatchIterator
from theanolm.training import Trainer, create_optimizer, CrossEntropyCost, \
                              NCECost, BlackoutCost
from theanolm.scoring import TextScorer
from theanolm.vocabulary import compute_word_counts

def add_arguments(parser):
    """Specifies the command line arguments supported by the "theanolm train"
    command.

    :type parser: argparse.ArgumentParser
    :param parser: a command line argument parser
    """

    argument_group = parser.add_argument_group("data")
    argument_group.add_argument(
        'model_path', metavar='MODEL-FILE', type=str,
        help='path where the best model state will be saved in HDF5 binary '
             'data format')
    argument_group.add_argument(
        '--training-set', metavar='FILE', type=TextFileType('r'), nargs='+',
        required=True,
        help='text files containing training data (UTF-8, one sentence per '
             'line, assumed to be compressed if the name ends in ".gz")')
    argument_group.add_argument(
        '--validation-file', metavar='VALID-FILE', type=TextFileType('r'),
        default=None,
        help='text file containing validation data for early stopping (UTF-8, '
             'one sentence per line, assumed to be compressed if the name ends '
             'in ".gz")')

    argument_group = parser.add_argument_group("vocabulary")
    argument_group.add_argument(
        '--vocabulary', metavar='FILE', type=str, default=None,
        help='word or class vocabulary to be used in the neural network input '
             'and output, in the format specified by the --vocabulary-format '
             'argument (UTF-8 text, default is to use all the words from the '
             'training data)')
    argument_group.add_argument(
        '--vocabulary-format', metavar='FORMAT', type=str, default='words',
        choices=['words', 'classes', 'srilm-classes'],
        help='format of the file specified with --vocabulary argument, one of '
             '"words" (one word per line, default), "classes" (word and class '
             'ID per line), "srilm-classes" (class name, membership '
             'probability, and word per line)')
    argument_group.add_argument(
        '--num-classes', metavar='N', type=int, default=None,
        help='generate N classes using a simple word frequency based algorithm '
             'when --vocabulary argument is not given (default is to not use '
             'word classes)')

    argument_group = parser.add_argument_group("network architecture")
    argument_group.add_argument(
        '--architecture', metavar='FILE', type=str, default='lstm300',
        help='path to neural network architecture description, or a standard '
             'architecture name, "lstm300" or "lstm1500" (default "lstm300")')

    argument_group = parser.add_argument_group("training process")
    argument_group.add_argument(
        '--sampling', metavar='FRACTION', type=float, nargs='*', default=[],
        help='randomly sample only FRACTION of each training file on each '
             'epoch (list the fractions in the same order as the training '
             'files)')
    argument_group.add_argument(
        '--sequence-length', metavar='N', type=int, default=100,
        help='ignore sentences longer than N words (default 100)')
    argument_group.add_argument(
        '--batch-size', metavar='N', type=int, default=16,
        help='each mini-batch will contain N sentences (default 16)')
    argument_group.add_argument(
        '--validation-frequency', metavar='N', type=int, default='5',
        help='cross-validate for reducing learning rate or early stopping N '
             'times per training epoch (default 5)')
    argument_group.add_argument(
        '--patience', metavar='N', type=int, default=4,
        help='allow perplexity to increase N consecutive cross-validations, '
             'before decreasing learning rate; if less than zero, never '
             'decrease learning rate (default 4)')
    argument_group.add_argument(
        '--random-seed', metavar='N', type=int, default=None,
        help='seed to initialize the random state (default is to seed from a '
             'random source provided by the oprating system)')

    argument_group = parser.add_argument_group("optimization")
    argument_group.add_argument(
        '--optimization-method', metavar='NAME', type=str, default='adagrad',
        choices=['sgd', 'nesterov', 'adagrad', 'adadelta', 'rmsprop-sgd',
                 'rmsprop-nesterov', 'adam'],
        help='optimization method, one of "sgd", "nesterov", "adagrad", '
             '"adadelta", "rmsprop-sgd", "rmsprop-nesterov", "adam" '
             '(default "adagrad")')
    argument_group.add_argument(
        '--learning-rate', metavar='ALPHA', type=float, default=0.1,
        help='initial learning rate (default 0.1)')
    argument_group.add_argument(
        '--l1-regularization', metavar='LAMBDA', type=float, default=None,
        help='add L1 regularization term with weight LAMBDA to the cost')
    argument_group.add_argument(
        '--l2-regularization', metavar='LAMBDA', type=float, default=None,
        help='add L2 regularization term with weight LAMBDA to the cost')
    argument_group.add_argument(
        '--momentum', metavar='BETA', type=float, default=0.9,
        help='momentum coefficient for momentum optimization methods (default '
             '0.9)')
    argument_group.add_argument(
        '--gradient-decay-rate', metavar='GAMMA', type=float, default=0.9,
        help='geometric rate for averaging gradients (default 0.9)')
    argument_group.add_argument(
        '--sqr-gradient-decay-rate', metavar='GAMMA', type=float, default=0.999,
        help='geometric rate for averaging squared gradients in Adam optimizer '
             '(default 0.999)')
    argument_group.add_argument(
        '--numerical-stability-term', metavar='EPSILON', type=float,
        default=1e-6,
        help='a value that is used to prevent instability when dividing by '
             'very small numbers (default 1e-6)')
    argument_group.add_argument(
        '--gradient-normalization', metavar='THRESHOLD', type=float,
        default=5,
        help='scale down the gradients if necessary to make sure their norm '
             '(normalized by mini-batch size) will not exceed THRESHOLD '
             '(default 5)')
    argument_group.add_argument(
        '--cost', metavar='NAME', type=str, default='cross-entropy',
        choices=['cross-entropy', 'nce', 'blackout'],
        help='cost function, one of "cross-entropy" (default), "nce" '
             '(noise-contrastive estimation), or "blackout"')
    argument_group.add_argument(
        '--num-noise-samples', metavar='K', type=int, default=5,
        help='sampling based costs sample K noise words per one training word '
             '(default 5)')
    argument_group.add_argument(
        '--noise-distribution', metavar='DIST', type=str, default='uniform',
        choices=['uniform', 'log-uniform', 'unigram'],
        help='sample noise from DIST; one of "uniform" (default, but less '
             'accurate), "log-uniform" (the vocabulary should be ordered by '
             'decreasing frequency), "unigram" (unigram distribution of words '
             'in training data, slow)')
    argument_group.add_argument(
        '--noise-dampening', metavar='ALPHA', type=float, default=0.5,
        help='the empirical unigram distribution is raised to the power ALPHA '
             'before sampling noise words; 0.0 corresponds to the uniform '
             'distribution and 1.0 corresponds to the unigram distribution '
             '(only applicable with --noise-distribution=unigram, default 0.5)')
    argument_group.add_argument(
        '--noise-sharing', metavar='SHARING', type=str, default=None,
        choices=['seq', 'batch', None],
        help='can be "seq" for sharing noise samples between mini-batch '
             'sequences, or "batch" for sharing noise samples across einter '
             'mini-batch for improved speed (default is no sharing, which is '
             'very slow)')
    argument_group.add_argument(
        '--exclude-unk', action="store_true",
        help="exclude <unk> tokens from cost and perplexity computations")
    argument_group.add_argument(
        '--weights', metavar='LAMBDA', type=float, nargs='*', default=[],
        help='scale a mini-batch update by LAMBDA if the data is from the '
             'corresponding training file (list the weights in the same order '
             'as the training files)')

    argument_group = parser.add_argument_group("early stopping")
    argument_group.add_argument(
        '--stopping-criterion', metavar='NAME', type=str,
        default='annealing-count',
        choices=['epoch-count', 'no-improvement', 'annealing-count'],
        help='selects a criterion for early-stopping, one of "epoch-count" '
             '(fixed number of epochs), "no-improvement" (no improvement since '
             'learning rate was decreased), "annealing-count" (default, '
             'learning rate is decreased a fixed number of times)')
    argument_group.add_argument(
        '--min-epochs', metavar='N', type=int, default=1,
        help='perform at least N training epochs (default 1)')
    argument_group.add_argument(
        '--max-epochs', metavar='N', type=int, default=100,
        help='perform at most N training epochs (default 100)')
    argument_group.add_argument(
        '--max-annealing-count', metavar='N', type=int, default=0,
        help='when using annealing-count stopping criterion, continue training '
             'after decreasing learning rate at most N times (default 0)')

    argument_group = parser.add_argument_group("configuration")
    argument_group.add_argument(
        '--default-device', metavar='DEVICE', type=str, default=None,
        help='when multiple GPUs are present, use DEVICE as default')

    argument_group = parser.add_argument_group("logging and debugging")
    argument_group.add_argument(
        '--log-file', metavar='FILE', type=str, default='-',
        help='path where to write log file (default is standard output)')
    argument_group.add_argument(
        '--log-level', metavar='LEVEL', type=str, default='info',
        choices=['debug', 'info', 'warn'],
        help='minimum level of events to log, one of "debug", "info", "warn" '
             '(default "info")')
    argument_group.add_argument(
        '--log-interval', metavar='N', type=int, default=1000,
        help='print statistics of every Nth mini-batch update; quiet if less '
             'than one (default 1000)')
    argument_group.add_argument(
        '--debug', action="store_true",
        help='use test values to get better error messages from Theano')
    argument_group.add_argument(
        '--print-graph', action="store_true",
        help='print Theano computation graph')
    argument_group.add_argument(
        '--profile', action="store_true",
        help='enable profiling Theano functions')

def _read_vocabulary(args, state):
    """If ``state`` contains data, reads the vocabulary from the HDF5 state.
    Otherwise reads a vocabulary file or constructs the vocabulary from the
    training set and writes it to the HDF5 state.

    If the state does not contain data and --vocabulary argument is given, reads
    the vocabulary from the file given after the argument. The rest of the words
    in the training set will be added as out-of-shortlist words.

    If the state does not contain data and no vocabulary is given, constructs a
    vocabulary that contains all the training set words. In that case,
    --num-classes argument can be used to control the number of classes.

    :type args: argparse.Namespace
    :param args: a collection of command line arguments

    :type state: hdf5.File
    :param state: HDF5 file where the vocabulary should be saved

    :rtype: Vocabulary
    :returns: the created vocabulary
    """

    if state.keys():
        logging.info("Reading vocabulary from existing network state.")
        result = Vocabulary.from_state(state)
        if not result.has_unigram_probs():
            # This is for backward compatibility. Remove at some point.
            logging.info("Computing unigram word probabilities from training "
                         "set.")
            word_counts = compute_word_counts(args.training_set)
            shortlist_words = list(result.id_to_word)
            shortlist_set = set(shortlist_words)
            oos_words = [x for x in word_counts.keys()
                         if x not in shortlist_set]
            result.id_to_word = numpy.asarray(shortlist_words + oos_words,
                                              dtype=object)
            result.word_to_id = {word: word_id
                                 for word_id, word in enumerate(result.id_to_word)}
            result.compute_probs(word_counts, update_class_probs=False)
            result.get_state(state)

    elif args.vocabulary is None:
        logging.info("Constructing vocabulary from training set.")
        word_counts = compute_word_counts(args.training_set)
        result = Vocabulary.from_word_counts(word_counts, args.num_classes)
        result.get_state(state)

    else:
        logging.info("Reading vocabulary from %s.", args.vocabulary)
        word_counts = compute_word_counts(args.training_set)
        oos_words = word_counts.keys()
        with open(args.vocabulary, 'rt', encoding='utf-8') as vocab_file:
            result = Vocabulary.from_file(vocab_file,
                                          args.vocabulary_format,
                                          oos_words=oos_words)

        if args.vocabulary_format == 'classes':
            logging.info("Computing class membership probabilities and unigram "
                         "probabilities for out-of-shortlist words.")
            update_class_probs = True
        else:
            logging.info("Computing unigram probabilities for out-of-shortlist "
                         "words.")
            update_class_probs = False
        result.compute_probs(word_counts,
                             update_class_probs=update_class_probs)
        result.get_state(state)

    logging.info("Number of words in vocabulary: %d",
                 result.num_words())
    logging.info("Number of words in shortlist: %d",
                 result.num_shortlist_words())
    logging.info("Number of word classes: %d",
                 result.num_classes())
    return result

def log_options(training_options, optimization_options, args):
    """Writes the command line arguments to debug log.
    """

    logging.debug("Training options:")
    for option_name, option_value in sorted(training_options.items()):
        logging.debug("  %s: %s", option_name, str(option_value))
    logging.debug("Optimization options:")
    for option_name, option_value in sorted(optimization_options.items()):
        logging.debug("  %s=%s", option_name, str(option_value))
    logging.debug("  cost_function=%s", args.cost)
    logging.debug("  noise_distribution=%s", args.noise_distribution)
    logging.debug("  noise_dampening=%d", args.noise_dampening)
    logging.debug("  noise_sharing=%s", args.noise_sharing
                                        if args.noise_sharing is not None
                                        else 'no')
    logging.debug("  exclude_unk=%s", 'yes' if args.exclude_unk else 'no')
    logging.debug("  l1_regularization=%f", args.l1_regularization
                                            if args.l1_regularization is not None
                                            else 0.0)
    logging.debug("  l2_regularization=%f", args.l2_regularization
                                            if args.l2_regularization is not None
                                            else 0.0)
    logging.debug("Data sampling: %s", str(numpy.array(args.sampling)))

def train(args):
    """A function that performs the "theanolm train" command.

    :type args: argparse.Namespace
    :param args: a collection of command line arguments
    """

    numpy.random.seed(args.random_seed)

    log_file = args.log_file
    log_level = getattr(logging, args.log_level.upper(), None)
    if not isinstance(log_level, int):
        print("Invalid logging level requested:", args.log_level)
        sys.exit(1)
    log_format = '%(asctime)s %(funcName)s: %(message)s'
    if args.log_file == '-':
        logging.basicConfig(stream=sys.stdout, format=log_format, level=log_level)
    else:
        logging.basicConfig(filename=log_file, format=log_format, level=log_level)

    if args.debug:
        theano.config.compute_test_value = 'warn'
        logging.info("Enabled computing test values for tensor variables.")
        logging.warning("GpuArray backend will fail random number generation!")
    else:
        theano.config.compute_test_value = 'off'
    theano.config.profile = args.profile
    theano.config.profile_memory = args.profile

    with h5py.File(args.model_path, 'a', driver='core') as state:
        vocabulary = _read_vocabulary(args, state)

        if args.num_noise_samples > vocabulary.num_classes():
            print("Number of noise samples ({}) is larger than the number of "
                  "classes. This doesn't make sense and would cause unigram "
                  "sampling to fail.".format(args.num_noise_samples))
            sys.exit(1)

        num_training_files = len(args.training_set)
        if len(args.weights) > num_training_files:
            print("You specified more weights than training files.")
            sys.exit(1)
        weights = numpy.ones(num_training_files).astype(theano.config.floatX)
        for index, weight in enumerate(args.weights):
            weights[index] = weight
        if len(args.sampling) > num_training_files:
            print("You specified more sampling coefficients than training "
                  "files.")
            sys.exit(1)

        training_options = {
            'batch_size': args.batch_size,
            'sequence_length': args.sequence_length,
            'validation_frequency': args.validation_frequency,
            'patience': args.patience,
            'stopping_criterion': args.stopping_criterion,
            'max_epochs': args.max_epochs,
            'min_epochs': args.min_epochs,
            'max_annealing_count': args.max_annealing_count
        }
        optimization_options = {
            'method': args.optimization_method,
            'epsilon': args.numerical_stability_term,
            'gradient_decay_rate': args.gradient_decay_rate,
            'sqr_gradient_decay_rate': args.sqr_gradient_decay_rate,
            'learning_rate': args.learning_rate,
            'weights': weights,
            'momentum': args.momentum,
            'max_gradient_norm': args.gradient_normalization,
            'num_noise_samples': args.num_noise_samples,
            'noise_sharing': args.noise_sharing,
        }

        log_options(training_options, optimization_options, args)

        logging.info("Creating trainer.")
        trainer = Trainer(training_options, vocabulary, args.training_set,
                          args.sampling)
        trainer.set_logging(args.log_interval)

        logging.info("Building neural network.")
        if args.architecture == 'lstm300' or args.architecture == 'lstm1500':
            architecture = Architecture.from_package(args.architecture)
        else:
            with open(args.architecture, 'rt', encoding='utf-8') as arch_file:
                architecture = Architecture.from_description(arch_file)

        default_device = get_default_device(args.default_device)
        network = Network(architecture, vocabulary, trainer.class_prior_probs,
                          default_device=default_device,
                          profile=args.profile)

        network.set_sampling(args.noise_distribution, args.noise_dampening,
                             args.noise_sharing)

        logging.info("Building optimizer.")
        exclude_id = vocabulary.word_to_id['<unk>'] if args.exclude_unk \
                     else None
        epsilon = args.numerical_stability_term
        if args.cost == 'cross-entropy':
            cost_function = CrossEntropyCost(network, exclude_id,
                                             args.l1_regularization,
                                             args.l2_regularization, epsilon)
        elif args.cost == 'nce':
            cost_function = NCECost(network, exclude_id, args.l1_regularization,
                                    args.l2_regularization, epsilon)
        else:
            assert args.cost == 'blackout'
            cost_function = BlackoutCost(network, exclude_id,
                                         args.l1_regularization,
                                         args.l2_regularization, epsilon)
        try:
            optimizer = create_optimizer(optimization_options, network,
                                         cost_function, profile=args.profile)
        except theano.gradient.DisconnectedInputError as e:
            print("Cannot train the neural network because some of the "
                  "parameters are disconnected from the output. Make sure all "
                  "the layers are correctly connected in the network "
                  "architecture. The error message was: `{}´".format(e))

        if args.print_graph:
            print("Cost function computation graph:")
            theano.printing.debugprint(optimizer.gradient_update_function)

        trainer.initialize(network, state, optimizer)

        if args.validation_file is not None:
            logging.info("Building text scorer for cross-validation.")
            scorer = TextScorer(network, use_shortlist=True,
                                exclude_unk=args.exclude_unk,
                                profile=args.profile)
            logging.info("Validation text: %s", args.validation_file.name)
            validation_mmap = mmap.mmap(args.validation_file.fileno(),
                                        0,
                                        prot=mmap.PROT_READ)
            validation_iter = \
                LinearBatchIterator(validation_mmap,
                                    vocabulary,
                                    batch_size=args.batch_size,
                                    max_sequence_length=args.sequence_length,
                                    map_oos_to_unk=False)
            trainer.set_validation(validation_iter, scorer)
        else:
            logging.info("Cross-validation will not be performed.")
            validation_iter = None

        logging.info("Training neural network.")
        trainer.train()

        if 'layers' not in state.keys():
            print("The model has not been trained. No cross-validations were "
                  "performed or training did not improve the model.")
        elif validation_iter is not None:
            network.set_state(state)
            perplexity = scorer.compute_perplexity(validation_iter)
            print("Best validation set perplexity:", perplexity)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A package that provides functionality for training and applying neural
network language models.
"""

from theanolm.vocabulary import Vocabulary
from theanolm.network import Network, Architecture, RecurrentState
from theanolm.scoring import TextScorer
from theanolm.textsampler import TextSampler
from theanolm.version import __version__



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the Vocabulary class.
"""

import logging

import numpy
import h5py
import theano

from theanolm.backend import IncompatibleStateError, InputError
from theanolm.vocabulary.wordclass import WordClass
from theanolm.parsing import utterance_from_line

def _add_special_tokens(id_to_word, word_id_to_class_id, word_classes):
    """Makes sure that the special symbols ``<s>``, ``</s>``, and ``<unk>``
    exist in the word list ``id_to_word``. If not, creates them and their word
    classes.

    :type id_to_word: list of strs
    :param id_to_word: mapping from word IDs to word names

    :type word_id_to_class_id: list of ints
    :param word_id_to_class_id: mapping from word IDs to indices in
                                ``word_classes``

    :type word_classes: list of WordClass objects
    :param word_classes: list of all the word classes
    """

    if len(id_to_word) != len(word_id_to_class_id):
        raise ValueError("Every word must be assigned to a class before adding "
                         "special tokens.")

    if '<s>' not in id_to_word:
        word_id = len(id_to_word)
        assert word_id == len(word_id_to_class_id)
        class_id = len(word_classes)
        id_to_word.append('<s>')
        word_id_to_class_id.append(class_id)
        word_class = WordClass(class_id, word_id, 1.0)
        word_classes.append(word_class)

    if '</s>' not in id_to_word:
        word_id = len(id_to_word)
        assert word_id == len(word_id_to_class_id)
        class_id = len(word_classes)
        id_to_word.append('</s>')
        word_id_to_class_id.append(class_id)
        word_class = WordClass(class_id, word_id, 1.0)
        word_classes.append(word_class)

    if '<unk>' not in id_to_word:
        word_id = len(id_to_word)
        assert word_id == len(word_id_to_class_id)
        class_id = len(word_classes)
        id_to_word.append('<unk>')
        word_id_to_class_id.append(class_id)
        word_class = WordClass(class_id, word_id, 1.0)
        word_classes.append(word_class)

class Vocabulary(object):
    """Word or Class Vocabulary

    Vocabulary class provides a mapping between the words and word IDs, as well
    as mapping from word IDs to class IDs. All shortlist words are mapped to a
    class. When classes are not used, the classes contain one word each. A
    vocabulary may also contain IDs for out-of-shortlist words. They are not
    mapped to a class and they are not predicted by the neural network.
    """

    def __init__(self, id_to_word, word_id_to_class_id, word_classes):
        """Constructs a vocabulary based on given word-to-class mapping.

        Expects the special tokens ``<s>``, ``</s>``, and ``<unk>`` to exist in
        the word list

        ``id_to_word`` defines the ID (continuous index) of every word.
        ``word_id_to_class_id`` defines the classes of shortlist words only. If
        ``id_to_word`` is longer that ``word_id_to_class_id``, the rest of the
        words will be considered out-of-shortlist.

        :type id_to_word: list of strs
        :param id_to_word: mapping from word IDs to word names

        :type word_id_to_class_id: list of ints
        :param word_id_to_class_id: mapping from word IDs to indices in
                                    ``word_classes``

        :type word_classes: list of WordClass objects
        :param word_classes: list of all the word classes
        """

        if not ({'<s>', '</s>', '<unk>'} <= set(id_to_word)):
            raise ValueError("Trying to construct shortlist vocabulary without "
                             "the special tokens <s>, </s>, and <unk>.")

        index = len(word_classes) - 1
        while index >= 0:
            word_class = word_classes[index]
            if len(word_class) == 1:
                word_id, _ = next(iter(word_class))
                if id_to_word[word_id].startswith('<'):
                    index -= 1
                    continue
            break
        self.num_normal_classes = index + 1

        for word_class in word_classes:
            word_class.normalize_probs()

        self.id_to_word = numpy.asarray(id_to_word, dtype=object)
        self.word_id_to_class_id = numpy.asarray(word_id_to_class_id)
        self._word_classes = numpy.asarray(word_classes)
        self.word_to_id = {word: word_id
                           for word_id, word in enumerate(self.id_to_word)}
        self._unigram_probs = None

    @classmethod
    def from_file(cls, input_file, input_format, oos_words=None):
        """Reads the shortlist words and possibly word classes from a vocabulary
        file.

        ``input_format`` is one of:

        * "words": ``input_file`` contains one word per line. Each word will be
                   assigned to its own class.
        * "classes": ``input_file`` contains a word followed by whitespace
                     followed by class ID on each line. Each word will be
                     assigned to the specified class. The class IDs can be
                     anything; they will be translated to consecutive numbers
                     after reading the file.
        * "srilm-classes": ``input_file`` contains a class name, membership
                           probability, and word, separated by whitespace, on
                           each line.

        The words read from the vocabulary file are put in the shortlist. If
        ``oos_words`` is given, those words are given an ID and added to the
        vocabulary as out-of-shortlist words if they don't exist in the
        vocabulary file.

        :type input_file: file object
        :param input_file: input vocabulary file

        :type input_format: str
        :param input_format: format of the input vocabulary file, "words",
                             "classes", or "srilm-classes"

        :type oos_words: list of strs
        :param oos_words: add words from this list to the vocabulary as
                          out-of-shortlist words, if they're not in the
                          vocabulary file
        """

        # We have also a set of the words just for faster checking if a word has
        # already been encountered.
        words = set()
        id_to_word = []
        word_id_to_class_id = []
        word_classes = []
        # Mapping from the IDs in the file to our internal class IDs.
        file_id_to_class_id = dict()

        for line in input_file:
            line = line.strip()
            fields = line.split()
            if not fields:
                continue
            if input_format == 'words' and len(fields) == 1:
                word = fields[0]
                file_id = None
                prob = 1.0
            elif input_format == 'classes' and len(fields) == 2:
                word = fields[0]
                file_id = int(fields[1])
                prob = 1.0
            elif input_format == 'srilm-classes' and len(fields) == 3:
                file_id = fields[0]
                prob = float(fields[1])
                word = fields[2]
            else:
                raise InputError("%d fields on one line of vocabulary file: %s"
                                 % (len(fields), line))

            if word in words:
                raise InputError("Word `%s´ appears more than once in the "
                                 "vocabulary file." % word)
            words.add(word)
            word_id = len(id_to_word)
            id_to_word.append(word)

            if file_id in file_id_to_class_id:
                class_id = file_id_to_class_id[file_id]
                word_classes[class_id].add(word_id, prob)
            else:
                # No ID in the file or a new ID.
                class_id = len(word_classes)
                word_class = WordClass(class_id, word_id, prob)
                word_classes.append(word_class)
                if file_id is not None:
                    file_id_to_class_id[file_id] = class_id

            assert word_id == len(word_id_to_class_id)
            word_id_to_class_id.append(class_id)

        _add_special_tokens(id_to_word, word_id_to_class_id, word_classes)
        words |= {'<s>', '</s>', '<unk>'}

        if oos_words is not None:
            for word in oos_words:
                if word not in words:
                    words.add(word)
                    id_to_word.append(word)

        return cls(id_to_word, word_id_to_class_id, word_classes)

    @classmethod
    def from_word_counts(cls, word_counts, num_classes=None):
        """Creates a vocabulary and classes from word counts. All the words will
        be in the shortlist.

        If ``num_classes`` is specified, words will be assigned to classes using
        modulo arithmetic. The class membership probabilities will be computed
        from the word counts.

        :type word_counts: dict
        :param word_counts: dictionary from words to the number of occurrences
                            in the corpus

        :type num_classes: int
        :param num_classes: number of classes to create in addition to the
                            special classes, or None for one class per word
        """

        # The special tokens should not be included when creating the classes.
        # They are added to separate classes in the end.
        word_counts = dict(word_counts)
        if '<s>' in word_counts:
            del word_counts['<s>']
        if '</s>' in word_counts:
            del word_counts['</s>']
        if '<unk>' in word_counts:
            del word_counts['<unk>']

        id_to_word = []
        word_id_to_class_id = []
        word_classes = []

        if num_classes is None:
            num_classes = len(word_counts)

        class_id = 0
        for word, _ in sorted(word_counts.items(),
                              key=lambda x: x[1]):
            word_id = len(id_to_word)
            id_to_word.append(word)

            if class_id < len(word_classes):
                word_classes[class_id].add(word_id, 1.0)
            else:
                assert class_id == len(word_classes)
                word_class = WordClass(class_id, word_id, 1.0)
                word_classes.append(word_class)

            assert word_id == len(word_id_to_class_id)
            word_id_to_class_id.append(class_id)
            class_id = (class_id + 1) % num_classes

        _add_special_tokens(id_to_word, word_id_to_class_id, word_classes)

        result = cls(id_to_word, word_id_to_class_id, word_classes)
        result.compute_probs(word_counts, update_class_probs=True)
        return result

    @classmethod
    def from_state(cls, state):
        """Reads the vocabulary from a network state.

        :type state: hdf5.File
        :param state: HDF5 file that contains the architecture parameters
        """

        if 'vocabulary' not in state:
            raise IncompatibleStateError(
                "Vocabulary is missing from neural network state.")
        h5_vocabulary = state['vocabulary']

        if 'words' not in h5_vocabulary:
            raise IncompatibleStateError(
                "Vocabulary parameter 'words' is missing from neural network "
                "state.")
        id_to_word = h5_vocabulary['words'].value

        if 'classes' not in h5_vocabulary:
            raise IncompatibleStateError(
                "Vocabulary parameter 'classes' is missing from neural network "
                "state.")
        word_id_to_class_id = h5_vocabulary['classes'].value

        if 'probs' not in h5_vocabulary:
            raise IncompatibleStateError(
                "Vocabulary parameter 'probs' is missing from neural network "
                "state.")
        num_classes = word_id_to_class_id.max() + 1
        word_classes = [None] * num_classes
        h5_probs = h5_vocabulary['probs'].value
        for word_id, prob in enumerate(h5_probs):
            class_id = word_id_to_class_id[word_id]
            if word_classes[class_id] is None:
                word_class = WordClass(class_id, word_id, prob)
                word_classes[class_id] = word_class
            else:
                word_classes[class_id].add(word_id, prob)

        result = cls(id_to_word.tolist(),
                     word_id_to_class_id.tolist(),
                     word_classes)

        if 'unigram_probs' in h5_vocabulary:
            result._unigram_probs = h5_vocabulary['unigram_probs'].value
            if len(result._unigram_probs) != result.num_words():
                raise IncompatibleStateError(
                    "Incorrect number of word unigram probabilities in neural "
                    "network state.")
            oos_probs = result._unigram_probs[result.num_shortlist_words():]
            if oos_probs.size:
                logging.debug("Out-of-shortlist word log probabilities are in "
                              "the range [%f, %f].",
                              numpy.log(oos_probs.min()),
                              numpy.log(oos_probs.max()))
        else:
            logging.debug("Word unigram probabilities are missing from state.")

        return result

    def compute_probs(self, word_counts, update_class_probs=True):
        """Computes word unigram probabilities and possibly recomputes class
        membership probabilities from word counts.

        Word unigram probabilities are computed for all words, excluding the
        start of sentence token.

        Class membership probabilities are updates only if
        ``update_class_probs`` is ``True`` and for classes whose words occur in
        ``word_counts``.

        Ensures that special tokens will always have nonzero class membership
        probabilities.

        :type word_counts: dict
        :param words_counts: mapping from word strings to counts

        :type update_class_probs: bool
        :param input_files: input text files
        """

        sos_id = self.word_to_id['<s>']
        eos_id = self.word_to_id['</s>']
        unk_id = self.word_to_id['<unk>']

        counts = numpy.zeros(self.num_words(), dtype='int64')
        for word, count in word_counts.items():
            if word in self.word_to_id:
                word_id = self.word_to_id[word]
                counts[word_id] = count

        self._unigram_probs = counts.astype(theano.config.floatX)
        total = self._unigram_probs.sum()
        if total > 0:
            self._unigram_probs /= total
        oos_probs = self._unigram_probs[self.num_shortlist_words():]
        if oos_probs.size:
            logging.debug("Out-of-shortlist word log probabilities are in the "
                          "range [%f, %f].",
                          numpy.log(oos_probs.min()),
                          numpy.log(oos_probs.max()))

        if not update_class_probs:
            return

        counts[sos_id] = max(counts[sos_id], 1)
        counts[eos_id] = max(counts[eos_id], 1)
        counts[unk_id] = max(counts[unk_id], 1)

        for cls in self._word_classes:
            cls_counts = dict()
            for word_id, _ in cls:
                cls_counts[word_id] = counts[word_id]
            cls_total = sum(cls_counts.values())
            if cls_total > 0:
                for word_id, count in cls_counts.items():
                    cls.set_prob(word_id, float(count) / cls_total)
            else:
                prob = 1.0 / len(cls)
                for word_id, _ in cls:
                    cls.set_prob(word_id, prob)

    def get_state(self, state):
        """Saves the vocabulary in a network state file.

        If there already is a vocabulary in the state, it will be replaced, so
        it has to have the same number of words.

        :type state: h5py.File
        :param state: HDF5 file for storing the neural network parameters
        """

        h5_vocabulary = state.require_group('vocabulary')

        if 'words' in h5_vocabulary:
            # XXX For backward compatibility. Remove at some point. XXX
            del h5_vocabulary['words']
            str_dtype = h5py.special_dtype(vlen=str)
            h5_vocabulary.create_dataset('words',
                                         data=self.id_to_word,
                                         dtype=str_dtype)
#            h5_vocabulary['words'][:] = self.id_to_word
        else:
            str_dtype = h5py.special_dtype(vlen=str)
            h5_vocabulary.create_dataset('words',
                                         data=self.id_to_word,
                                         dtype=str_dtype)

        if 'classes' in h5_vocabulary:
            h5_vocabulary['classes'][:] = self.word_id_to_class_id
        else:
            h5_vocabulary.create_dataset('classes',
                                         data=self.word_id_to_class_id)

        probs = [self._word_classes[class_id].get_prob(word_id)
                 for word_id, class_id in enumerate(self.word_id_to_class_id)]
        if 'probs' in h5_vocabulary:
            h5_vocabulary['probs'][:] = probs
        else:
            h5_vocabulary.create_dataset('probs', data=probs)

        if self.has_unigram_probs():
            if 'unigram_probs' in h5_vocabulary:
                h5_vocabulary['unigram_probs'][:] = self._unigram_probs
            else:
                h5_vocabulary.create_dataset('unigram_probs',
                                             data=self._unigram_probs)

    def num_words(self):
        """Returns the number of words in the vocabulary. This includes
        shortlist and out-of-shortlist words.

        :rtype: int
        :returns: the number of words in the shirtlist
        """

        return self.id_to_word.size

    def num_shortlist_words(self):
        """Returns the number of words in the shortlist. Only the shortlist
        words are assigned to a class and are predicted by the neural network.

        :rtype: int
        :returns: the number of words in the shortlist
        """

        return self.word_id_to_class_id.size

    def num_classes(self):
        """Returns the number of word classes.

        :rtype: int
        :returns: the number of words classes
        """

        return self._word_classes.size

    def words_to_ids(self, words):
        """Translates words into word IDs.

        :type words: list of strs
        :param words: a list of words

        :rtype: ndarray
        :returns: the given words translated into word IDs
        """

        unk_id = self.word_to_id['<unk>']
        result = numpy.zeros(len(words), dtype='int64')
        for index, word in enumerate(words):
            if word in self.word_to_id:
                result[index] = self.word_to_id[word]
            else:
                result[index] = unk_id
        return result

    def class_ids_to_word_ids(self, class_ids):
        """Samples a word from the membership probability distribution of a
        class. (If classes are not used, returns the one word in the class.)

        :type class_ids: list of ints
        :param class_ids: list of class IDs

        :rtype: list of ints
        :returns: a word ID from each of the given classes
        """

        return [self._word_classes[class_id].sample()
                for class_id in class_ids]

    def get_word_prob(self, word_id):
        """Returns the class membership probability of a word.

        :type word_id: int
        :param word_id: ID of a word

        :rtype: float
        :returns: the probability of the word within its class
        """

        class_id = self.word_id_to_class_id[word_id]
        word_class = self._word_classes[class_id]
        return word_class.get_prob(word_id)

    def get_class_memberships(self, word_ids):
        """Finds the classes and class membership probabilities given a matrix
        of word IDs.

        Word IDs that are not in the shortlist are converted to ``<unk>``.

        :type word_ids: ndarray
        :param word_ids: a matrix containing word IDs

        :rtype: tuple of ndarrays
        :returns: two matrices, the first one containing class IDs and the
                  second one containing class membership probabilities
        """

        unk_id = self.word_to_id['<unk>']
        word_ids = numpy.copy(word_ids)
        word_ids[word_ids >= self.num_shortlist_words()] = unk_id
        class_ids = self.word_id_to_class_id[word_ids]
        word_classes = self._word_classes[class_ids]
        get_probs = numpy.vectorize(lambda wc, wid: wc.get_prob(wid))
        return class_ids, get_probs(word_classes, word_ids)

    def words(self):
        """A generator for iterating through the words in the vocabulary.

        :rtype: generator for str
        :returns: generates the next word in the vocabulary
        """

        for word in self.word_to_id.keys():
            yield word

    def in_shortlist(self, word_id):
        """Checks if the word with given ID is in the shortlist.

        :type word_id: int
        :param word_id: a word ID

        :rtype: bool
        :returns: ``True`` if the word is in shortlist, ``False`` otherwise
        """

        return word_id < self.word_id_to_class_id.size

    def has_unigram_probs(self):
        """Checks if the word unigram probabilities are computed and
        ``get_oos_probs()`` can be called.

        :rtype: bool
        :returns: ``True`` if the unigram probabilities are computed, ``False``
                  otherwise
        """

        return self._unigram_probs is not None

    def get_oos_probs(self):
        """Returns an array that can be indexed by word ID to obtain the
        probability of a word given that the word is an out-of-shortlist word.

        The probabilities predicted by the network should be multiplied by the
        corresponding elements of the returned array to obtain a correctly
        normalized probability distribution. The returned probabilities are one
        for shortlist words, meaning that the shortlist probabilities are not
        affected. For other words, the probability mass of ``<unk>`` is divided
        according to their unigram frequencies.

        :rtype: ndarray
        :returns: an array that maps a word ID to the log probability that
                  should be added to the log probability predicted by the
                  network
        """

        shortlist_size = self.num_shortlist_words()
        result = numpy.copy(self._unigram_probs)
        result[:shortlist_size] = 1.0
        total_oos_prob = result[shortlist_size:].sum()
        result[shortlist_size:] /= total_oos_prob
        return result

    def get_class_probs(self):
        """Computes the class probabilities by summing the member word
        probabilities. Normalizes the resulting probability distribution.

        :rtype: ndarray
        :returns: probability distribution of the word classes
        """

        shortlist_size = self.num_shortlist_words()
        class_ids = self.word_id_to_class_id[:shortlist_size]
        probs = self._unigram_probs[:shortlist_size]
        result = numpy.zeros(self.num_classes(), dtype=theano.config.floatX)
        numpy.add.at(result, class_ids, probs)

        total = result.sum()
        if total > 0:
            result /= total

        return result

    def __contains__(self, word):
        """Tests if ``word`` is included in the vocabulary.

        :type word: str
        :param word: a word

        :rtype: bool
        :returns: True if ``word`` is in the vocabulary, False otherwise.
        """

        return word in self.word_to_id



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the TextScorer class.
"""

import logging

import numpy
import theano
import theano.tensor as tensor

from theanolm.backend import NumberError
from theanolm.backend import test_value
from theanolm.parsing import utterance_from_line

class TextScorer(object):
    """Text Scoring Using a Neural Network Language Model
    """

    def __init__(self, network, use_shortlist=True, exclude_unk=False,
                 profile=False):
        """Creates two Theano function, ``self._target_logprobs_function()``,
        which computes the log probabilities predicted by the neural network for
        the words in a mini-batch, and ``self._total_logprob_function()``, which
        returns the total log probability.

        Both functions take as arguments four matrices:

        1. Word IDs in the shape of a mini-batch. The functions will only use
           the input words (not the last time step).
        2. Class IDs in the shape of a mini-batch. The functions will slice this
           into input and output.
        3. Class membership probabilities in the shape of a mini-batch, but only
           for the output words (not the first time step).
        4. Mask in the shape of a mini-batch, but only for the output words (not
           for the first time step).

        ``self._target_logprobs_function()`` will return a matrix of predicted
        log probabilities for the output words (excluding the first time step)
        and the mask. ``<unk>`` tokens are also masked out if ``exclude_unk`` is
        set to ``True``. ``self._total_logprob_function()`` will return the
        total log probability of the predicted (unmasked) words and the number
        of those words.

        :type network: Network
        :param network: the neural network object

        :type use_shortlist: bool
        :param use_shortlist: if ``True``, the ``<unk>`` probability is
                              distributed among the out-of-shortlist words

        :type exclude_unk: bool
        :param exclude_unk: if set to ``True``, ``<unk>`` tokens are excluded
                            from probability computation

        :type profile: bool
        :param profile: if set to True, creates a Theano profile object
        """

        self._vocabulary = network.vocabulary
        self._unk_id = self._vocabulary.word_to_id['<unk>']

        # The functions take as input a mini-batch of word IDs and class IDs,
        # and slice input and target IDs for the network.
        batch_word_ids = tensor.matrix('textscorer/batch_word_ids',
                                       dtype='int64')
        batch_word_ids.tag.test_value = test_value(
            size=(21, 4), high=self._vocabulary.num_words())
        batch_class_ids = tensor.matrix('textscorer/batch_class_ids',
                                        dtype='int64')
        batch_class_ids.tag.test_value = test_value(
            size=(21, 4), high=self._vocabulary.num_classes())
        membership_probs = tensor.matrix('textscorer/membership_probs',
                                         dtype=theano.config.floatX)
        membership_probs.tag.test_value = test_value(
            size=(20, 4), high=1.0)

        # Convert out-of-shortlist words to <unk> in input.
        shortlist_size = self._vocabulary.num_shortlist_words()
        input_word_ids = batch_word_ids[:-1]
        oos_indices = tensor.ge(input_word_ids, shortlist_size).nonzero()
        input_word_ids = tensor.set_subtensor(input_word_ids[oos_indices],
                                              self._unk_id)
        # Out-of-shortlist words are already in <unk> class, because they don't
        # have own classes.
        input_class_ids = batch_class_ids[:-1]
        target_class_ids = batch_class_ids[1:]
        # Target word IDs are not used by the network. We need them to compute
        # probabilities for out-of-shortlist word.
        target_word_ids = batch_word_ids[1:]

        logprobs = tensor.log(network.target_probs())
        # Add logprobs from the class membership of the predicted word.
        logprobs += tensor.log(membership_probs)

        mask = network.mask
        if use_shortlist and network.oos_logprobs is not None:
            # The probability of out-of-shortlist words (which is the <unk>
            # probability) is multiplied by the fraction of the actual word
            # within the set of OOS words.
            logprobs += network.oos_logprobs[target_word_ids]
            # Always exclude OOV words when using a shortlist - No probability
            # mass is left for them.
            mask *= tensor.neq(target_word_ids, self._unk_id)
        elif exclude_unk:
            # If requested, ignore OOS and OOV probabilities.
            mask *= tensor.neq(target_word_ids, self._unk_id)
            mask *= tensor.lt(target_word_ids, shortlist_size)

        # Ignore unused input variables, because is_training is only used by
        # dropout layer.
        masked_logprobs = logprobs * tensor.cast(mask, theano.config.floatX)
        self._target_logprobs_function = theano.function(
            [batch_word_ids, batch_class_ids, membership_probs, network.mask],
            [masked_logprobs, mask],
            givens=[(network.input_word_ids, input_word_ids),
                    (network.input_class_ids, input_class_ids),
                    (network.target_class_ids, target_class_ids),
                    (network.is_training, numpy.int8(0))],
            name='target_logprobs',
            on_unused_input='ignore',
            profile=profile)

        # If some word is not in the training data, its class membership
        # probability will be zero. We want to ignore those words. Multiplying
        # by the mask is not possible, because those logprobs will be -inf.
        mask *= tensor.neq(membership_probs, 0.0)
        masked_logprobs = tensor.switch(mask, logprobs, 0.0)
        self._total_logprob_function = theano.function(
            [batch_word_ids, batch_class_ids, membership_probs, network.mask],
            [masked_logprobs.sum(), mask.sum()],
            givens=[(network.input_word_ids, input_word_ids),
                    (network.input_class_ids, input_class_ids),
                    (network.target_class_ids, target_class_ids),
                    (network.is_training, numpy.int8(0))],
            name='total_logprob',
            on_unused_input='ignore',
            profile=profile)

        # These are updated by score_line().
        self.num_words = 0
        self.num_unks = 0

    def score_batch(self, word_ids, class_ids, membership_probs, mask):
        """Computes the log probabilities predicted by the neural network for
        the words in a mini-batch.

        The result will be returned in a list of lists. The indices will be a
        transpose of those of the input matrices, so that the first index is the
        sequence, not the time step. The lists will contain ``None`` values in
        place of any ``<unk>`` tokens, if the constructor was given
        ``exclude_unk=True``. When using a shortlist, the lists will always
        contain ``None`` in place of OOV words, and if ``exclude_unk=True`` was
        given, also in place of OOS words. Words with zero class membership
        probability will have ``-inf`` log probability.

        :type word_ids: numpy.ndarray of an integer type
        :param word_ids: a 2-dimensional matrix, indexed by time step and
                         sequence, that contains the word IDs

        :type class_ids: numpy.ndarray of an integer type
        :param class_ids: a 2-dimensional matrix, indexed by time step and
                          sequence, that contains the class IDs

        :type membership_probs: numpy.ndarray of a floating point type
        :param membership_probs: a 2-dimensional matrix, indexed by time step
                                 and sequences, that contains the class
                                 membership probabilities of the words

        :type mask: numpy.ndarray of a floating point type
        :param mask: a 2-dimensional matrix, indexed by time step and sequence,
                     that masks out elements past the sequence ends

        :rtype: list of lists
        :returns: logprob of each word in each sequence, ``None`` values
                  indicating excluded <unk> tokens
        """

        result = []
        membership_probs = membership_probs.astype(theano.config.floatX)

        # target_logprobs_function() uses the word and class IDs of the entire
        # mini-batch, but membership probs and mask are only for the output.
        logprobs, new_mask = self._target_logprobs_function(word_ids,
                                                            class_ids,
                                                            membership_probs[1:],
                                                            mask[1:])
        for seq_index in range(logprobs.shape[1]):
            seq_mask = mask[1:, seq_index]
            seq_logprobs = logprobs[seq_mask == 1, seq_index]
            # The new mask also masks excluded tokens, replace those with None.
            seq_mask = new_mask[seq_mask == 1, seq_index]
            seq_logprobs = [lp if m == 1 else None
                            for lp, m in zip(seq_logprobs, seq_mask)]
            result.append(seq_logprobs)

        return result

    def compute_perplexity(self, batch_iter):
        """Computes the perplexity of text read using the given iterator.

        ``batch_iter`` is an iterator to the input data. On each call it creates
        a two 2-dimensional matrices, both indexed by time step and sequence.
        The first matrix contains the word IDs, the second one masks out
        elements past the sequence ends.

        ``<unk>`` tokens will be excluded from the perplexity computation, if
        the constructor was given ``exclude_unk=True``. When using a shortlist,
        OOV words are always excluded, and if ``exclude_unk=True`` was given,
        OOS words are also excluded. Words with zero class membership
        probability are always excluded.

        :type batch_iter: BatchIterator
        :param batch_iter: an iterator that creates mini-batches from the input
                           data

        :rtype: float
        :returns: perplexity, i.e. exponent of negative log probability
                  normalized by the number of words
        """

        logprob = 0
        num_words = 0

        for word_ids, _, mask in batch_iter:
            class_ids, membership_probs = \
                self._vocabulary.get_class_memberships(word_ids)
            membership_probs = membership_probs.astype(theano.config.floatX)

            # total_logprob_function() uses the word and class IDs of the entire
            # mini-batch, but membership probs and mask are only for the output.
            batch_logprob, batch_num_words = \
                self._total_logprob_function(word_ids,
                                             class_ids,
                                             membership_probs[1:],
                                             mask[1:])
            if numpy.isnan(batch_logprob):
                self._debug_log_batch(word_ids, class_ids, membership_probs, mask)
                raise NumberError("Log probability of a mini-batch is NaN.")
            if numpy.isneginf(batch_logprob):
                self._debug_log_batch(word_ids, class_ids, membership_probs, mask)
                raise NumberError("Probability of a mini-batch is zero.")
            if batch_logprob > 0.0:
                self._debug_log_batch(word_ids, class_ids, membership_probs, mask)
                raise NumberError("Probability of a mini-batch is greater than one.")

            logprob += batch_logprob
            num_words += batch_num_words

        if num_words == 0:
            raise ValueError("Zero words for computing perplexity. Does the "
                             "evaluation data contain only OOV words?")
        cross_entropy = -logprob / num_words
        return numpy.exp(cross_entropy)

    def score_sequence(self, word_ids, class_ids, membership_probs):
        """Computes the log probability of a word sequence.

        ``<unk>`` tokens will be excluded from the probability computation, if
        the constructor was given ``exclude_unk=True``. When using a shortlist,
        OOV words are always excluded, and if ``exclude_unk=True`` was given,
        OOS words are also excluded. Words with zero class membership
        probability are always excluded.

        :type word_ids: ndarray
        :param word_ids: a vector of word IDs

        :type class_ids: list of ints
        :param class_ids: corresponding class IDs

        :type membership_probs: list of floats
        :param membership_probs: list of class membership probabilities

        :rtype: float
        :returns: log probability of the word sequence
        """

        # Create 2-dimensional matrices representing the transposes of the
        # vectors.
        word_ids = numpy.transpose(word_ids[numpy.newaxis])
        class_ids = numpy.array([[x] for x in class_ids], numpy.int64)
        membership_probs = numpy.array(
            [[x] for x in membership_probs]).astype(theano.config.floatX)
        # Mask used by the network is all ones.
        mask = numpy.ones(word_ids.shape, numpy.int8)

        # total_logprob_function() uses the word and class IDs of the entire
        # mini-batch, but membership probs and mask are only for the output.
        logprob, _ = self._total_logprob_function(word_ids,
                                                  class_ids,
                                                  membership_probs[1:],
                                                  mask[1:])
        if numpy.isnan(logprob):
            self._debug_log_batch(word_ids, class_ids, membership_probs, mask)
            raise NumberError("Log probability of a sequence is NaN.")
        if numpy.isneginf(logprob):
            self._debug_log_batch(word_ids, class_ids, membership_probs, mask)
            raise NumberError("Probability of a sequence is zero.")
        if logprob > 0.0:
            self._debug_log_batch(word_ids, class_ids, membership_probs, mask)
            raise NumberError("Probability of a sequence is greater than one.")

        return logprob

    def score_line(self, line, vocabulary):
        """Scores a line of text.

        Start-of-sentence and end-of-sentece tags (``<s>`` and ``</s>``) will be
        inserted at the beginning and the end of the line, if they're missing.
        If the line is empty, ``None`` will be returned, instead of interpreting
        it as the empty sentence ``<s> </s>``.

        ``<unk>`` tokens will be excluded from the probability computation, if
        the constructor was given ``exclude_unk=True``. When using a shortlist,
        OOV words are always excluded, and if ``exclude_unk=True`` was given,
        OOS words are also excluded. Words with zero class membership
        probability are always excluded.

        :type line: str
        :param line: a sequence of words

        :type vocabulary: Vocabulary
        :param vocabulary: vocabulary for converting the words to word IDs

        :rtype: float
        :returns: log probability of the word sequence, or None if the line is
                  empty
        """

        words = utterance_from_line(line)
        if not words:
            return None

        word_ids = vocabulary.words_to_ids(words)
        unk_id = vocabulary.word_to_id['<unk>']
        self.num_words += word_ids.size
        self.num_unks += numpy.count_nonzero(word_ids == unk_id)

        class_ids = [vocabulary.word_id_to_class_id[word_id]
                     for word_id in word_ids]
        probs = [vocabulary.get_word_prob(word_id)
                 for word_id in word_ids]

        return self.score_sequence(word_ids, class_ids, probs)

    def _debug_log_batch(self, word_ids, class_ids, membership_probs, mask):
        """Writes the target word IDs, their log probabilities, and the mask to
        the debug log.

        :type word_ids: numpy.ndarray of an integer type
        :param word_ids: a 2-dimensional matrix, indexed by time step and
                         sequence, that contains the word IDs

        :type class_ids: numpy.ndarray of an integer type
        :param class_ids: a 2-dimensional matrix, indexed by time step and
                          sequence, that contains the class IDs

        :type membership_probs: numpy.ndarray of a floating point type
        :param membership_probs: a 2-dimensional matrix, indexed by time step
                                 and sequences, that contains the class
                                 membership probabilities of the words

        :type mask: numpy.ndarray of a floating point type
        :param mask: a 2-dimensional matrix, indexed by time step and sequence,
                     that masks out elements past the sequence ends
        """

        membership_probs = membership_probs.astype(theano.config.floatX)

        # target_logprobs_function() uses the word and class IDs of the entire
        # mini-batch, but membership probs and mask are only for the output.
        logprobs, new_mask = self._target_logprobs_function(word_ids,
                                                            class_ids,
                                                            membership_probs[1:],
                                                            mask[1:])
        for seq_index in range(logprobs.shape[1]):
            target_word_ids = word_ids[1:, seq_index]
            seq_mask = mask[1:, seq_index]
            seq_word_ids = target_word_ids[seq_mask == 1]
            seq_logprobs = logprobs[seq_mask == 1, seq_index]
            # The new mask also masks excluded tokens.
            seq_mask = new_mask[seq_mask == 1, seq_index]
            logging.debug("Sequence %i target word IDs: [%s]",
                          seq_index, ", ".join(str(x) for x in seq_word_ids))
            logging.debug("Sequence %i mask: [%s]",
                          seq_index, ", ".join(str(x) for x in seq_mask))
            logging.debug("Sequence %i logprobs: [%s]",
                          seq_index, ", ".join(str(x) for x in seq_logprobs))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the KaldiLattice used by Kaldi lattice rescorer.
"""

from collections import namedtuple
import logging

from theanolm.backend import InputError
from theanolm.backend.probfunctions import logprob_type
from theanolm.scoring.lattice import Lattice

def read_kaldi_vocabulary(input_file):
    """Reads a word-to-ID mapping from a Kaldi vocabulary file.

    :type input_file: file object
    :param input_file: a Kaldi vocabulary file (words.txt)

    :rtype: dict
    :returns: a mapping from words to Kaldi word IDs
    """

    result = dict()
    for line in input_file:
        parts = line.split()
        if not parts:
            continue
        if len(parts) != 2:
            raise InputError("Invalid Kaldi vocabulary file.")
        word = parts[0]
        word_id = int(parts[1])
        result[word] = word_id
    return result

class KaldiLattice(Lattice):
    """Kaldi Lattice

    A word lattice that can be read in Kaldi CompactLattice Format
    """

    def __init__(self, lattice_lines, id_to_word):
        """Reads a Kaldi lattice file.

        If ``lattice_lines`` is ``None``, creates an empty lattice (useful for
        testing).

        :type lattice_lines: list of strs
        :param lattice_lines: list of lines in Kaldi CompactLattice text format

        :type id_to_word: list
        :param id_to_word: mapping of word IDs to words
        """

        super().__init__()

        # No logarithm base conversion.
        self._log_scale = logprob_type(1.0)

        self._initial_node_id = None
        self._final_node_id = None
        final_node = self.Node(None)
        final_node.final = True
        self.nodes = []

        if lattice_lines is None:
            self._num_nodes = 0
            self._num_links = 0
            return

        self.utterance_id = lattice_lines[0].strip()
        for line in lattice_lines[1:]:
            line = line.strip()
            parts = line.split()
            if not parts:
                continue
            if len(parts) == 1:
                state_to = kaldi_word_id = None
                str_weight = ""
            elif len(parts) == 2:
                state_to = kaldi_word_id = None
                str_weight = parts[1]
            elif len(parts) == 4:
                state_to = int(parts[1])
                kaldi_word_id = int(parts[2])
                str_weight = parts[3]
            else:
                raise InputError("Invalid number of fields in lattice `{}´ "
                                 "line `{}´.".format(self.utterance_id, line))
            state_from = int(parts[0])

            weight_parts = str_weight.split(',')
            graph_logprob = logprob_type(weight_parts[0]) \
                            if len(weight_parts) > 0 and weight_parts[0] \
                            else logprob_type(0.0)
            graph_logprob = -graph_logprob * self._log_scale
            ac_logprob = logprob_type(weight_parts[1]) \
                         if len(weight_parts) > 1 and weight_parts[1] \
                         else logprob_type(0.0)
            ac_logprob = -ac_logprob * self._log_scale
            transitions = weight_parts[2] if len(weight_parts) > 2 else ""

            if self._initial_node_id is None:
                self._initial_node_id = state_from

            for id in range(len(self.nodes), state_from + 1):
                self.nodes.append(self.Node(id))
            if state_to is not None:
                for id in range(len(self.nodes), state_to + 1):
                    self.nodes.append(self.Node(id))
                link = self._add_link(self.nodes[state_from], self.nodes[state_to])
                link.word = id_to_word[kaldi_word_id]
                if link.word == "<eps>":
                    link.word = None
                elif link.word == "#0":
                    raise InputError("Lattice `{}´ contains backoff transitions. "
                                     "Fix with Kaldi commands."
                                     .format(self.utterance_id))
                elif (link.word == "<s>") or (link.word == "</s>"):
                    raise InputError("Lattice `{}´ contains traditional start "
                                     "and end of sentence symbols."
                                     .format(self.utterance_id))
            else:
                link = self._add_link(self.nodes[state_from], final_node)
                link.word = None

            link.ac_logprob = ac_logprob
            link.lm_logprob = graph_logprob
            link.transitions = transitions

        if self._initial_node_id is None:
            raise InputError("No links in lattice `{}´."
                             .format(self.utterance_id))
        self.initial_node = self.nodes[self._initial_node_id]

        final_node.id = len(self.nodes)
        self.nodes.append(final_node)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the RMSProp optimizer with SGD.
"""

import numpy
import theano.tensor as tensor

from theanolm.backend import Parameters
from theanolm.training.basicoptimizer import BasicOptimizer

class RMSPropSGDOptimizer(BasicOptimizer):
    """RMSProp Variation of Stochastic Gradient Descent Optimization Method

    At the time of writing, RMSProp is an unpublished method. Usually people
    cite slide 29 of Lecture 6 of Geoff Hinton's Coursera class:
    http://www.cs.toronto.edu/~tijmen/csc321/slides/lecture_slides_lec6.pdf

    The idea is simply to maintain a running average of the squared gradient for
    each parameter, and divide the gradient by the root of the mean squared
    gradient (RMS). This makes RMSProp take steps near 1 whenever the gradient
    is of constant magnitude, and larger steps whenever the local scale of the
    gradient starts to increase.
    """

    def __init__(self, optimization_options, network, *args, **kwargs):
        """Creates an RMSProp SGD optimizer.

        :type optimization_options: dict
        :param optimization_options: a dictionary of optimization options

        :type network: Network
        :param network: the neural network object
        """

        self._params = Parameters()
        for path, param in network.get_variables().items():
            # Initialize mean squared gradient to ones, otherwise the first
            # update will be divided by close to zero.
            self._params.add(path + '_mean_sqr_gradient',
                             numpy.ones_like(param.get_value()))

        # geometric rate for averaging gradients
        if 'gradient_decay_rate' not in optimization_options:
            raise ValueError("Gradient decay rate is not given in training "
                             "options.")
        self._gamma = optimization_options['gradient_decay_rate']

        super().__init__(optimization_options, network, *args, **kwargs)

    def _get_param_updates(self, alpha):
        """Returns Theano expressions for updating the model parameters and any
        additional parameters required by the optimizer.

        :type alpha: Variable
        :param alpha: a scale to be applied to the model parameter updates

        :rtype: iterable over pairs (shared variable, new expression)
        :returns: expressions how to update the optimizer parameters
        """

        result = []
        deltas = dict()
        for path, gradient in zip(self.network.get_variables(),
                                  self._gradients):
            ms_gradient_old = self._params[path + '_mean_sqr_gradient']
            ms_gradient = \
                self._gamma * ms_gradient_old + \
                (1.0 - self._gamma) * tensor.sqr(gradient)
            result.append((ms_gradient_old, ms_gradient))

            rms_gradient = tensor.sqrt(ms_gradient + self._epsilon)
            deltas[path] = -gradient / rms_gradient
        self._normalize(deltas)

        for path, param_old in self.network.get_variables().items():
            delta = deltas[path]
            result.append((param_old, param_old + alpha * delta))
        return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the Stochastic Gradient Descent optimizer.
"""

import numpy

from theanolm.backend import Parameters
from theanolm.training.basicoptimizer import BasicOptimizer

class SGDOptimizer(BasicOptimizer):
    """Stochastic Gradient Descent Optimization Method
    """

    def __init__(self, optimization_options, network, *args, **kwargs):
        """Creates a Stochastic Gradient Descent optimizer. SGD optimizer does
        not use additional parameters.

        :type optimization_options: dict
        :param optimization_options: a dictionary of optimization options

        :type network: Network
        :param network: the neural network object
        """

        self._params = Parameters()

        super().__init__(optimization_options, network, *args, **kwargs)

    def _get_param_updates(self, alpha):
        """Returns Theano expressions for updating the model parameters and any
        additional parameters required by the optimizer.

        :type alpha: Variable
        :param alpha: a scale to be applied to the model parameter updates

        :rtype: iterable over pairs (shared variable, new expression)
        :returns: expressions how to update the optimizer parameters
        """

        deltas = dict()
        result = []
        for path, gradient in zip(self.network.get_variables(),
                                  self._gradients):
            deltas[path] = -gradient
        self._normalize(deltas)

        for path, param_old in self.network.get_variables().items():
            delta = deltas[path]
            result.append((param_old, param_old + alpha * delta))
        return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the Network class.
"""

from collections import OrderedDict
import sys
import logging

import h5py
import numpy
import theano
import theano.tensor as tensor
from theano.sandbox.rng_mrg import MRG_RandomStreams as RandomStreams

from theanolm import Vocabulary
from theanolm.backend import IncompatibleStateError, InputError
from theanolm.backend import UniformDistribution, LogUniformDistribution
from theanolm.backend import MultinomialDistribution
from theanolm.backend import test_value
from theanolm.network.architecture import Architecture
from theanolm.network.networkinput import NetworkInput
from theanolm.network.projectionlayer import ProjectionLayer
from theanolm.network.fullyconnectedlayer import FullyConnectedLayer
from theanolm.network.additionlayer import AdditionLayer
from theanolm.network.grulayer import GRULayer
from theanolm.network.lstmlayer import LSTMLayer
from theanolm.network.highwaylayer import HighwayLayer
from theanolm.network.glulayer import GLULayer
from theanolm.network.softmaxlayer import SoftmaxLayer
from theanolm.network.hsoftmaxlayer import HSoftmaxLayer
from theanolm.network.dropoutlayer import DropoutLayer
from theanolm.network.bidirectionallayer import BidirectionalLayer
from theanolm.network.samplingoutputlayer import SamplingOutputLayer

def create_layer(layer_options, *args, **kwargs):
    """Constructs one of the Layer classes based on a layer definition.

    :type layer_type: str
    :param layer_type: a text string describing the layer type
    """

    layer_type = layer_options['type']
    if layer_type == 'projection':
        return ProjectionLayer(layer_options, *args, **kwargs)
    elif layer_type == 'fc' or layer_type == 'tanh':
        return FullyConnectedLayer(layer_options, *args, **kwargs)
    elif layer_type == 'add':
        return AdditionLayer(layer_options, *args, **kwargs)
    elif layer_type == 'lstm':
        return LSTMLayer(layer_options, *args, **kwargs)
    elif layer_type == 'gru':
        return GRULayer(layer_options, *args, **kwargs)
    elif layer_type == 'blstm' or layer_type == 'bgru':
        return BidirectionalLayer(layer_options, *args, **kwargs)
    elif layer_type == 'highway' or layer_type == 'highwaytanh':
        return HighwayLayer(layer_options, *args, **kwargs)
    elif layer_type == 'glu':
        return GLULayer(layer_options, *args, **kwargs)
    elif layer_type == 'softmax':
        return SoftmaxLayer(layer_options, *args, **kwargs)
    elif layer_type == 'hsoftmax':
        return HSoftmaxLayer(layer_options, *args, **kwargs)
    elif layer_type == 'dropout':
        return DropoutLayer(layer_options, *args, **kwargs)
    else:
        raise ValueError("Invalid layer type requested: " + layer_type)

class Network(object):
    """Neural Network

    A class that creates the actual neural network graph using Theano. Functions
    that train and apply the neural network can be created by passing the input
    and output variables to ``theano.function()``.
    """

    class Mode():
        """Network Mode Selection

        Enumeration of options for selecting network mode. This will create a
        slightly different output for different purposes.

          - ``minibatch``: Process mini-batches with multiple sequences and time
                           steps. The output is a matrix with one less time
                           steps containing the probabilities of the words at
                           the next time step.
        """
        def __init__(self, minibatch=True, nce=False):
            self.minibatch = minibatch
            self.nce = nce

    def __init__(self, architecture, vocabulary, class_prior_probs=None,
                 mode=None, exclude_unk=False, default_device=None,
                 profile=False):
        """Initializes the neural network parameters for all layers, and
        creates Theano shared variables from them.

        :type architecture: Architecture
        :param architecture: an object that describes the network architecture

        :type vocabulary: Vocabulary
        :param vocabulary: mapping between word IDs and word classes

        :type class_prior_probs: numpy.ndarray
        :param class_prior_probs: empirical (unigram) distribution of the output
                                  classes (only required for training)

        :type mode: Network.Mode
        :param mode: selects mini-batch or single time step processing

        :type exclude_unk: bool
        :param exclude_unk: if set to ``True``, set ``<unk>`` probability to
                            zero before normalizing the network outputs
                            (required to get exact normalization during
                            inference)

        :type default_device: str
        :param default_device: default device where to store the shared variables

        :type profile: bool
        :param profile: if set to True, creates a Theano profile object
        """

        self.vocabulary = vocabulary
        self.architecture = architecture
        self.mode = self.Mode() if mode is None else mode

        M1 = 2147483647
        M2 = 2147462579
        random_seed = [
            numpy.random.randint(0, M1),
            numpy.random.randint(0, M1),
            numpy.random.randint(1, M1),
            numpy.random.randint(0, M2),
            numpy.random.randint(0, M2),
            numpy.random.randint(1, M2)]
        self.random = RandomStreams(random_seed)

        # Word and class inputs will be available to NetworkInput layers.
        self.input_word_ids = tensor.matrix('network/input_word_ids',
                                            dtype='int64')
        self.input_class_ids = tensor.matrix('network/input_class_ids',
                                             dtype='int64')
        if self.mode.minibatch:
            self.input_word_ids.tag.test_value = test_value(
                size=(20, 4), high=vocabulary.num_shortlist_words())
            self.input_class_ids.tag.test_value = test_value(
                size=(20, 4), high=vocabulary.num_classes())
        else:
            self.input_word_ids.tag.test_value = test_value(
                size=(1, 4), high=vocabulary.num_shortlist_words())
            self.input_class_ids.tag.test_value = test_value(
                size=(1, 4), high=vocabulary.num_classes())

        # Should the output layer set the <unk> probability to zero? In that
        # case OOVs are not counted when computing perplexity.
        self.exclude_unk = exclude_unk

        # Default device for shared variables.
        self._default_device = default_device

        # During training, the output layer bias vector is initialized to the
        # unigram probabilities.
        self.class_prior_probs = class_prior_probs

        # A shortlist model adds these logprobs to OOS logprobs predicted by the
        # network.
        if vocabulary.has_unigram_probs():
            oos_logprobs = numpy.log(vocabulary.get_oos_probs())
            oos_logprobs = oos_logprobs.astype(theano.config.floatX)
            if self._default_device is None:
                self.oos_logprobs = theano.shared(oos_logprobs,
                                                  'network/oos_logprobs')
            else:
                self.oos_logprobs = theano.shared(oos_logprobs,
                                                  'network/oos_logprobs',
                                                  target=self._default_device)
        else:
            self.oos_logprobs = None

        # Recurrent layers will create these lists, used to initialize state
        # variables of appropriate sizes, for doing forward passes one step at a
        # time.
        self.recurrent_state_input = []
        self.recurrent_state_size = []

        # Create the layers.
        logging.debug("Creating layers.")
        self.layers = OrderedDict()
        for input_options in architecture.inputs:
            network_input = NetworkInput(input_options, self)
            self.layers[network_input.name] = network_input
        for layer_description in architecture.layers:
            layer_options = self._layer_options_from_description(
                layer_description)
            if layer_options['name'] == architecture.output_layer:
                layer_options['size'] = vocabulary.num_classes()
            # 'devices' not in layer_options is for backward compatibility.
            # Remove at some point.
            if ('devices' not in layer_options) or (not layer_options['devices']):
                layer_options['devices'] = [default_device]
            layer = create_layer(layer_options, self, profile=profile)
            self.layers[layer.name] = layer
        self.output_layer = self.layers[architecture.output_layer]
        num_params = sum(layer.num_params() for layer in self.layers.values())
        logging.debug("Total number of model parameters: %d", num_params)

        # This list will be filled by the recurrent layers to contain the
        # recurrent state outputs, for doing forward passes one step at a time.
        self.recurrent_state_output = [None] * len(self.recurrent_state_size)

        # This input variable can be used to specify the classes whose
        # probabilities will be computed, instead of the whole distribution.
        self.target_class_ids = tensor.matrix('network/target_class_ids',
                                              dtype='int64')
        if self.mode.minibatch:
            self.target_class_ids.tag.test_value = test_value(
                size=(20, 4), high=vocabulary.num_classes())
        else:
            self.target_class_ids.tag.test_value = test_value(
                size=(1, 4), high=vocabulary.num_classes())

        # This input variable is used only for detecting <unk> target words.
        self.target_word_ids = tensor.matrix('network/target_word_ids',
                                             dtype='int64')
        if self.mode.minibatch:
            self.target_word_ids.tag.test_value = test_value(
                size=(20, 4), high=vocabulary.num_shortlist_words())
        else:
            self.target_word_ids.tag.test_value = test_value(
                size=(1, 4), high=vocabulary.num_shortlist_words())

        # mask is used to mask out the rest of the input matrix, when a sequence
        # is shorter than the maximum sequence length. The mask is kept as int8
        # data type, which is how Tensor stores booleans.
        if self.mode.minibatch:
            self.mask = tensor.matrix('network/mask', dtype='int8')
            self.mask.tag.test_value = test_value(size=(20, 4), high=True)
        else:
            self.mask = tensor.ones(self.input_word_ids.shape, dtype='int8')

        # Dropout layer needs to know whether we are training or evaluating.
        self.is_training = tensor.scalar('network/is_training', dtype='int8')
        self.is_training.tag.test_value = 1

        # num_noise_samples tells sampling based methods how many noise classes
        # to sample.
        self.num_noise_samples = tensor.scalar('network/num_noise_samples',
                                               dtype='int64')
        self.num_noise_samples.tag.test_value = 3
        self.noise_distribution = None

        for layer in self.layers.values():
            layer.create_structure()

    @classmethod
    def from_file(cls, model_path, mode=None, exclude_unk=False,
                  default_device=None):
        """Reads a model from an HDF5 file.

        :type model_path: str
        :param model_path: path to a HDF5 model file

        :type mode: Network.Mode
        :param mode: selects mini-batch or single time step processing

        :type exclude_unk: bool
        :param exclude_unk: if set to ``True``, set ``<unk>`` probability to
                            zero before normalizing the network outputs
                            (required to get exact normalization during
                            inference)

        :type default_device: str
        :param default_device: default device where to store the shared variables
        """

        with h5py.File(model_path, 'r') as state:
            logging.info("Reading vocabulary from network state.")
            #sys.stdout.flush()
            vocabulary = Vocabulary.from_state(state)
            logging.info("Number of words in vocabulary: {}"
                         .format(vocabulary.num_words()))
            logging.info("Number of words in shortlist: {}"
                         .format(vocabulary.num_shortlist_words()))
            logging.info("Number of word classes: {}"
                         .format(vocabulary.num_classes()))
            logging.info("Building neural network.")
            #sys.stdout.flush()
            architecture = Architecture.from_state(state)
            result = cls(architecture, vocabulary, mode=mode,
                         exclude_unk=exclude_unk, default_device=default_device)
            logging.info("Restoring neural network state.")
            result.set_state(state)
            return result

    def set_sampling(self, type, dampening, sharing):
        """Defines sampling of noise words for sampling-based output.

        When using a sampling based output layer, it needs to know the prior
        distribution of the classes, and how many words to sample. The number of
        noise classes per training word is controlled by the
        ``num_noise_samples`` tensor variable. The type of the prior
        distribution is defined by the ``type`` argument. If it is 'unigram',
        the unigram distribution of the classes in the training data is raised
        to the power of ``dampening`` and saved in a shared variable, so that we
        don't have to transfer the vector on every call of a Theano function.

        :type type: str
        :param type: either 'uniform', 'log-uniform' (logarithm of the samples
                     is uniformly distributed), or 'unigram' (unigram
                     distribution of the words raised to the power of
                     ``dampening``)

        :type dampening: float
        :param dampening: exponent to which unigram distribution is raised,
                          effectively dampening it

        :type sharing: str
        :param sharing: either ``None`` for k samples per mini-batch element,
                        'seq' for k samples per time step, or 'batch' for k
                        samples in total
        """

        output_layer = self.output_layer

        if (type == 'uniform') or ((type == 'unigram') and (dampening == 0.0)):
            distribution = UniformDistribution(self.random,
                                               self.vocabulary.num_classes())
        elif type == 'log-uniform':
            distribution = LogUniformDistribution(self.random,
                                                  self.vocabulary.num_classes())
        elif type == 'unigram':
            probs = numpy.power(self.class_prior_probs, dampening)
            probs /= probs.sum()
            probs = probs.astype(theano.config.floatX)
            if self._default_device is None:
                probs = theano.shared(probs, 'network/noise_probs')
            else:
                probs = theano.shared(probs, 'network/noise_probs',
                                      target=self._default_device)
            distribution = MultinomialDistribution(self.random, probs)
        else:
            raise ValueError("Invalid noise distribution requested: `{}´"
                             .format(type))

        if not isinstance(output_layer, SamplingOutputLayer):
            self._noise_sample = None
            self._noise_sample_logprobs = None
        elif sharing is None:
            self._noise_sample, self._noise_sample_logprobs = \
                output_layer.get_sample_tensors(distribution)
        elif sharing == 'seq':
            self._noise_sample, self._noise_sample_logprobs = \
                output_layer.get_seqshared_sample_tensors(distribution)
        elif sharing == 'batch':
            self._noise_sample, self._noise_sample_logprobs = \
                output_layer.get_shared_sample_tensors(distribution)
        else:
            raise ValueError("Unknown noise sample sharing: `{}´"
                             .format(sharing))

        self.noise_distribution = distribution

    def get_state(self, state):
        """Pulls parameter values from Theano shared variables.

        If there already is a parameter in the state, it will be replaced, so it
        has to have the same number of elements.

        :type state: h5py.File
        :param state: HDF5 file for storing the neural network parameters
        """

        for layer in self.layers.values():
            layer.get_state(state)

        self.architecture.get_state(state)

    def set_state(self, state):
        """Sets the values of Theano shared variables.

        Requires that ``state`` contains values for all the neural network
        parameters.

        :type state: h5py.File
        :param state: HDF5 file that contains the neural network parameters
        """

        for layer in self.layers.values():
            layer.set_state(state)

        try:
            self.architecture.check_state(state)
        except IncompatibleStateError as error:
            raise IncompatibleStateError(
                "Attempting to restore state of a network that is incompatible "
                "with this architecture. " + str(error))

    def get_variables(self):
        """Returns a dictionary of the model parameters.

        The returned dictionary is a mapping from path strings that identify the
        parameters to symbolic tensors. The parameter tensors are used by
        optimizers to compute gradients with regards to the parameters. They may
        also be used to create optimization parameters that are specific to
        model parameters, and compute the L1 or L2 regularization.

        :rtype: dict
        :returns: mapping from parameter path to symbolic tensors
        """

        result = dict()
        for layer in self.layers.values():
            result.update(layer.get_variables())
        return result

    def add_recurrent_state(self, size):
        """Adds a recurrent state variable and returns its index.

        Used by recurrent layers to add a state variable that has to be passed
        from one time step to the next, when generating text or computing
        lattice probabilities.

        :type size: int
        :param size: size of the state vector

        :rtype size: int
        :param size: index of the new recurrent state variable
        """

        index = len(self.recurrent_state_size)
        assert index == len(self.recurrent_state_input)

        # The variables are in the structure of a mini-batch (3-dimensional
        # array) to keep the layer functions general.
        variable = tensor.tensor3('network/recurrent_state_' + str(index),
                                  dtype=theano.config.floatX)
        variable.tag.test_value = test_value(size=(1, 4, size), high=1.0)

        self.recurrent_state_size.append(size)
        self.recurrent_state_input.append(variable)

        return index

    def output_probs(self):
        """Returns the output probabilities for the whole vocabulary.

        :rtype: Variable
        :returns: a symbolic 3-dimensional matrix that contains a probability
                  for each time step, each sequence, and each output class
        """

        if not hasattr(self.output_layer, 'output_probs'):
            raise RuntimeError("The final layer is not an output layer.")
        return self.output_layer.output_probs

    def target_probs(self):
        """Returns the output probabilities for the predicted words.

        Can be used only when target_class_ids is given.

        :rtype: Variable
        :returns: a symbolic 2-dimensional matrix that contains the target word
                  probability for each time step and each sequence
        """

        if not hasattr(self.output_layer, 'target_probs'):
            raise RuntimeError("The final layer is not an output layer.")
        return self.output_layer.target_probs

    def unnormalized_logprobs(self):
        """Returns the unnormalized log probabilities for the predicted words.

        These are the preactivations of the output layer, before softmax. As the
        softmax output is exponential, these can be seen as the unnormalized log
        probabilities.

        Only computed when target_class_ids is given and using softmax output.

        :rtype: Variable
        :returns: a symbolic 2-dimensional matrix that contains the unnormalized
                  target word probability for each time step and each sequence
        """

        if not hasattr(self.output_layer, 'unnormalized_logprobs'):
            raise RuntimeError("The final layer is not a softmax layer, and "
                               "unnormalized probabilities are needed.")
        return self.output_layer.unnormalized_logprobs

    def noise_sample(self):
        """Returns the classes sampled from a noise distribution, and their log
        probabilities.

        Only computed when target_class_ids is given and using softmax output.

        If ``sharing`` is None``, the sample is a 3-dimensional matrix of k
        class IDs for each time step and sequence. If ``sharing`` is 'seq', the
        sample is a 2-dimensional matrix, k class IDs for each time step. If
        ``sharing`` is 'batch', the sample is a vector of k class IDs in total.
        The log probabilities are always returned in a 3-dimensional matrix, as
        they differ for each time step and sequence.

        :rtype: tuple of two Variables
        :returns: noise class IDs and their log probabilities
        """

        if (self._noise_sample is None) or \
           (self._noise_sample_logprobs is None):
            raise InputError(
                "The output layer does not support sampling noise words. "
                "Sampling-based costs can be used only with normal softmax and "
                "only one device.")

        return self._noise_sample, self._noise_sample_logprobs

    def _layer_options_from_description(self, description):
        """Creates layer options based on textual architecture description.

        Most of the fields in a layer description are kept as strings. The field
        ``input_layers`` is converted to a list of actual layers found from
        ``self.layers``.

        :type description: dict
        :param description: dictionary of textual layer fields

        :rtype: dict
        :result: layer options
        """

        result = dict()
        for variable, value in description.items():
            if variable == 'inputs':
                try:
                    result['input_layers'] = [self.layers[x] for x in value]
                except KeyError as e:
                    raise InputError("Input layer `{}´ does not exist, when "
                                     "creating layer `{}´."
                                     .format(e.args[0], description['name']))
            else:
                result[variable] = value
        return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the SamplingOutputLayer class, a base class for
output layer that implements sampling.
"""

from abc import ABCMeta

import numpy
import theano.tensor as tensor

from theanolm.network.basiclayer import BasicLayer


class SamplingOutputLayer(BasicLayer, metaclass=ABCMeta):
    """Sampling Support for Output Layer

    Base class for output layers with support for sampling noise words and
    computing unnormalized probabilities. This is needed for Noise-contrastive
    estimation and BlackOut.
    """

    def _get_unnormalized_logprobs(self):
        """Creates tensor variable that computes unnormalized output
        (preactivations).

        Unnormalized probabilities and noise probabilities are used by noise-
        contrastive estimation. Softmax output is exponential, so the
        preactivations can be seen as unnormalized log probabilities.

        Sampling-based output is not supported when using multiple GPUs.

        :rtype: symbolic 2-dimensional tensor
        :returns: the unnormalized log probabilities of the target words for
                  each sequence in each time step, or ``None`` if using multiple
                  GPUs.
        """

        layer_input = self._layer_input
        if layer_input is None:
            raise RuntimeError("Unnormalized outputs requested before "
                               "constructing the network structure.")
        if len(self._devices) > 1:
            return None

        num_time_steps = layer_input.shape[0]
        num_sequences = layer_input.shape[1]
        target_class_ids = self._network.target_class_ids[:, :, None]
        result = self._get_target_preact(layer_input, target_class_ids)
        return result.reshape([num_time_steps, num_sequences])

    def get_sample_tensors(self, noise_distribution):
        """Creates tensor variables for sampling k unique noise words per
        mini-batch element for NCE and BlackOut.

        Sampling-based output is not supported when using multiple GPUs.

        :type noise_distribution: ClassDistribution
        :param noise_distribution: defines the distribution where the noise
                                   words should be drawn from

        :rtype: tuple of two symbolic 3-dimensional tensors
        :returns: the k sampled class IDs and their log probabilities for each
                  time step in each sequence, or ``(None, None)`` if using
                  multiple GPUs.
        """

        layer_input = self._layer_input
        if layer_input is None:
            raise RuntimeError("Sampling-based output requested before "
                               "constructing the network structure.")
        if len(self._devices) > 1:
            return None, None

        num_time_steps = layer_input.shape[0]
        num_sequences = layer_input.shape[1]
        num_samples = self._network.num_noise_samples
        num_classes = numpy.int64(self._network.vocabulary.num_classes())

        minibatch_size = num_time_steps * num_sequences
        sample = noise_distribution.sample(minibatch_size, num_samples)
        sample = sample.reshape([num_time_steps, num_sequences, num_samples])
        return sample, self._get_target_preact(layer_input, sample)

    def get_seqshared_sample_tensors(self, noise_distribution):
        """Creates tensor variables for sampling noise for NCE and BlackOut.
        Creates k samples for each time step. These are shared across the
        sequences in the mini-batch.

        Sampling-based output is not supported when using multiple GPUs.

        :type noise_distribution: ClassDistribution
        :param noise_distribution: defines the distribution where the noise
                                   words should be drawn from

        :rtype: tuple of two symbolic 3-dimensional tensors
        :returns: the k sampled class IDs for each time step (the second
                  dimension being empty), and their log probabilities for each
                  time step in each sequence, or ``(None, None)`` if using
                  multiple GPUs.
        """

        layer_input = self._layer_input
        if layer_input is None:
            raise RuntimeError("Sampling-based output requested before "
                               "constructing the network structure.")
        if len(self._devices) > 1:
            return None, None

        num_time_steps = layer_input.shape[0]
        num_samples = self._network.num_noise_samples
        num_batch_samples = num_time_steps * num_samples
        num_classes = numpy.int64(self._network.vocabulary.num_classes())

        # Sampling k noise words per time step is inefficient with multinomial.
        sample = noise_distribution.sample(1, num_batch_samples)
        sample = sample.reshape([num_time_steps, num_samples])
        return sample[:, None, :], \
               self._get_target_seq_preact(layer_input, sample)

    def get_shared_sample_tensors(self, noise_distribution):
        """Creates tensor variables for sampling noise for NCE and BlackOut.
        Creates k samples for each time step. These are shared across the
        sequences in the mini-batch.

        Sampling-based output is not supported when using multiple GPUs.

        :type noise_distribution: ClassDistribution
        :param noise_distribution: defines the distribution where the noise
                                   words should be drawn from

        :rtype: tuple of two symbolic 3-dimensional tensors
        :returns: the k sampled class IDs (the first and second dimension being
                  empty) and their log probabilities for each time step in each
                  sequence, or ``(None, None)`` if using multiple GPUs.
        """

        layer_input = self._layer_input
        if layer_input is None:
            raise RuntimeError("Sampling-based output requested before "
                               "constructing the network structure.")
        if len(self._devices) > 1:
            return None, None

        num_samples = self._network.num_noise_samples
        num_classes = numpy.int64(self._network.vocabulary.num_classes())

        # Sample k noise words in total. These are shared across mini-batch.
        sample = noise_distribution.sample(1, num_samples)
        sample = sample[0, :]
        return sample[None, None, :], \
               self._get_target_list_preact(layer_input, sample)

    def _get_target_preact(self, layer_input, target_class_ids):
        """Constructs the preactivations for given targets. One or more target
        outputs are given for each element in the mini-batch.

        :type layer_input: Variable
        :param layer_input: a 3-dimensional tensor that contains the input
                            vector for each time step in each sequence

        :type target_class_ids: Variable
        :param target_class_ids: a 3-dimensional tensor that contains one or
                                 more target class IDs for each time step in
                                 each sequence

        :rtype: Variable
        :returns: a 3-dimensional tensor that contains the preactivation for
                  each target class, for each time step in each sequence
        """

        weight = self._params[self._param_path('input/W')]
        bias = self._params[self._param_path('input/b')]
        weight = weight.T
        weight = weight[target_class_ids, :]
        # The old GPU backend does not implement GpuAdvancedIncSubtensor1_dev20
        # for vectors, which is why the very slow GpuAdvancedIncSubtensor1 will
        # be selected if we index a vector.
        bias = bias[:, None]
        bias = bias[target_class_ids, 0]
#        bias = bias[target_class_ids]
        return (layer_input[:, :, None, :] * weight).sum(3) + bias

    def _get_target_seq_preact(self, layer_input, target_class_ids):
        """Constructs the preactivations for given targets. One or more target
        outputs are given for each time step in a 2-dimensional matrix.

        :type layer_input: Variable
        :param layer_input: a 3-dimensional tensor that contains the input
                            vector for each time step in each sequence

        :type target_class_ids: Variable
        :param target_class_ids: a 2-dimensional tensor that contains one or
                                 more target class IDs for each time step

        :rtype: Variable
        :returns: a 3-dimensional tensor that contains the preactivation for
                  each target class, for each time step in each sequence
        """

        weight = self._params[self._param_path('input/W')]
        bias = self._params[self._param_path('input/b')]
        weight = weight.T
        weight = weight[target_class_ids, :]
        # The old GPU backend does not implement GpuAdvancedIncSubtensor1_dev20
        # for vectors, which is why the very slow GpuAdvancedIncSubtensor1 will
        # be selected if we index a vector.
        bias = bias[:, None]
        bias = bias[target_class_ids, 0]
#        bias = bias[target_class_ids]
        result = layer_input[:, :, None, :] * weight[:, None, :, :]
        result = result.sum(3)
        result += bias[:, None, :]
        return result

    def _get_target_list_preact(self, layer_input, target_class_ids):
        """Structures the preactivations for a list of target classes.
        Preactivations at each word are computed for all the targets.

        :type layer_input: Variable
        :param layer_input: a 3-dimensional tensor that contains the input
                            vector for each time step in each sequence

        :type target_class_ids: Variable
        :param target_class_ids: a list of target classes

        :rtype: Variable
        :returns: a 3-dimensional tensor that contains the preactivation for
                  every target word, at each time step of each sequence
        """

        weight = self._params[self._param_path('input/W')]
        bias = self._params[self._param_path('input/b')]
        weight = weight[:, target_class_ids]
        # The old GPU backend does not implement GpuAdvancedIncSubtensor1_dev20
        # for vectors, which is why the very slow GpuAdvancedIncSubtensor1 will
        # be selected if we index a vector.
        bias = bias[:, None]
        bias = bias[target_class_ids, 0]
#        bias = bias[target_class_ids]
        return tensor.dot(layer_input, weight) + bias



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the BasicLayer class, a base class for all layers.
"""

from abc import abstractmethod, ABCMeta
import logging

import theano.tensor as tensor

from theanolm.backend import Parameters, conv1d
from theanolm.network.weightfunctions import random_matrix, matrix_from_value

class BasicLayer(object, metaclass=ABCMeta):
    """Superclass for Neural Network Layers
    """

    def __init__(self, layer_options, network, profile=False):
        """Saves some attributes that are common to all layers.

        :type layer_options: dict
        :param layer_options: dictionary of layer options

        :type network: Network
        :param network: the network object creating this layer

        :type profile: bool
        :param profile: if set to True, creates a Theano profile object
        """

        self.name = layer_options['name']
        self._input_layers = layer_options['input_layers']
        self._params = Parameters()
        self._devices = layer_options['devices']

        if 'size' in layer_options:
            self.output_size = int(layer_options['size'])
        else:
            self.output_size = \
                sum([x.output_size for x in self._input_layers])

        if 'activation' in layer_options:
            activation = layer_options['activation']
        else:
            activation = 'tanh'
        if activation == 'tanh':
            self._activation = tensor.tanh
        elif activation == 'relu':
            self._activation = tensor.nnet.relu
        elif activation == 'leakyrelu':
            self._activation = lambda x: tensor.nnet.relu(x, 0.3)
        else:
            raise InputError("Unsupported activation function: {}"
                             .format(activation))

        if 'reverse_time' in layer_options:
            self._reverse_time = bool(layer_options['reverse_time'])
        else:
            self._reverse_time = False

        logging.debug("- %s name=%s inputs=[%s] size=%d activation=%s%s "
                      "devices=[%s]",
                      self.__class__.__name__,
                      self.name,
                      ', '.join([x.name for x in self._input_layers]),
                      self.output_size,
                      activation,
                      ' reverse,' if self._reverse_time else '',
                      ', '.join([str(x) for x in self._devices]))

        self._network = network
        self._profile = profile

    @abstractmethod
    def create_structure(self):
        """Creates the symbolic graph of this layer.

        Sets self.output to a symbolic matrix that describes the output of this
        layer.
        """

        assert False

    def get_state(self, state):
        """Pulls parameter values from Theano shared variables.

        If there already is a parameter in the state, it will be replaced, so it
        has to have the same number of elements.

        :type state: h5py.File
        :param state: HDF5 file for storing the neural network parameters
        """

        self._params.get_state(state)

    def set_state(self, state):
        """Sets the values of Theano shared variables.

        :type state: h5py.File
        :param state: HDF5 file that contains the neural network parameters
        """

        self._params.set_state(state)

    def num_params(self):
        """Returns the number of parameters in this layer.

        This method is used just for reporting the number of parameters in the
        model. Normally there is just one set of parameters.

        :rtype: int
        :returns: the number of parameters used by the layer
        """

        return self._params.total_size

    def get_variables(self):
        """Returns a dictionary of the shared variables.

        This function is used by the optimizers to create optimization
        parameters that are specific to network parameters, and compute
        gradients with regard to the parameters. Normally there is just one set
        of parameters.

        :rtype: dict
        :returns: mapping from parameter path to Theano shared variables
        """

        return self._params.get_variables()

    def _param_path(self, param_name, device=None):
        """Returns the HDF5 path used to address a parameter.

        :type param_name: str
        :param param_name: name of a parameter within this layer

        :type device: str
        :param device: ``None`` for parameters that reside on the default device
                       only; otherwise returns the path used to address the part
                       of the parameter that resides on the given device

        :rtype: str
        :returns: full path of the parameter in a HDF5 file.
        """

        result = 'layers/' + self.name + '/' + param_name
        if device is not None:
            result += '/' + device
        return result

    def _get_param(self, param_name, device=None):
        """Returns a Theano tensor variable by parameter name.

        :type param_name: str
        :param param_name: name of a parameter within the layer

        :type device: str
        :param device: ``None`` for parameters that reside on the default device
                       only; otherwise returns the part of the parameter that
                       resides on the given device

        :rtype: Variable
        :returns: the corresponding tensor variable
        """

        return self._params[self._param_path(param_name, device)]

    def _init_weight(self, param_name, shape, scale=None, count=1,
                     split_to_devices=True):
        """Generates a weight matrix from “standard normal” distribution.

        If ``shape`` contains two dimensions that match, generates an orthogonal
        matrix. In that case scale is ignored. Orthogonal weights are useful for
        two reasons:

        1. Multiplying by an orthogonal weight preserves the norm of the
           input vector, which should help avoid exploding and vanishing
           gradients.
        2. The row and column vectors are orthonormal to one another, which
           should help avoid two vectors learning to produce the same features.

        If ``count`` is specified, creates a concatenation of several similar
        submatrices (same shape but different content).

        If ``split_to_devices`` is ``True``, splits the weight to equal parts on
        the last dimension, and creates one parameter for each device.

        :type shape: list or tuple of ints
        :param shape: sizes of the weight dimensions; normally the first one is
                      the dimensionality of the input data and the second one is
                      the dimensionality of the output data

        :type scale: float
        :param scale: if other than ``None``, the matrix will be scaled by this
                      factor, unless an orthogonal matrix is created

        :type count: int
        :param count: concatenate this many weight matrices with the same shape

        :type split_to_devices: bool
        :param split_to_devices: if ``True``, creates on every device a
                                 parameter that contains one part of the weight
        """

        path = self._param_path(param_name)
        weight = random_matrix(shape, scale, count)
        if not split_to_devices:
            self._params.add(path, weight)
        elif (len(self._devices) == 1) and (self._devices[0] is None):
            # This layer has not been assigned to a specific device.
            self._params.add(path, weight)
        else:
            split_ranges = self._split_per_device(weight.shape[-1])
            for device, split_range in zip(self._devices, split_ranges):
                assert device is not None
                self._params.add('{}/{}'.format(path, device),
                                 weight[..., split_range], device)

    def _init_bias(self, param_name, shape, value=None, split_to_devices=True):
        """Initializes a bias vector with given value.

        If ``value`` is not given, initializes the vector with zero value. If
        ``value`` is a list, creates a concatenation of as many vectors as there
        are elements in the list.

        If ``split_to_devices`` is ``True``, splits the array to equal parts on
        the last dimension, and creates one parameter for each device.

        :type param_name: str
        :param param_name: name for the parameter within the layer

        :type shape: int or tuple of ints
        :param shape: size of the vector, or a tuple of the sizes of each
                      dimension (in case ``value`` is a list, each part will
                      have this size)

        :type value: float, numpy.ndarray or list
        :param value: the value or array to initialize the elements to, or a
                      list of values or arrays to create a concatenation of
                      vectors

        :type split_to_devices: bool
        :param split_to_devices: if ``True``, creates on every device a
                                 parameter that contains one part of the array
        """

        path = self._param_path(param_name)
        bias = matrix_from_value(shape, value)
        if not split_to_devices:
            self._params.add(path, matrix_from_value(shape, value))
        elif (len(self._devices) == 1) and (self._devices[0] is None):
            # This layer has not been assigned to a specific device.
            self._params.add(path, matrix_from_value(shape, value))
        else:
            split_ranges = self._split_per_device(bias.shape[-1])
            for device, split_range in zip(self._devices, split_ranges):
                assert device is not None
                self._params.add('{}/{}'.format(path, device),
                                 bias[..., split_range], device)

    def _split_per_device(self, total_size):
        """Returns a list of range objects that divide ``total_size`` equally to
        as many parts as there are devices.

        :type total_size: int
        :param total_size: total size of a parameter

        :rtype: list of ranges
        :returns: ``total_size`` divided into as many parts as there are devices
                  assigned to this layer
        """

        num_devices = len(self._devices)
        if num_devices < 1:
            raise RuntimeError("No devices assigned to this layer.")
        if total_size < num_devices:
            raise ValueError("Cannot split matrix of size {} to {} devices."
                             .format(total_size, num_devices))

        result = []
        quotient, remainder = divmod(total_size, num_devices)
        start_index = 0
        for i in range(1, num_devices + 1):
            end_index = i * quotient + min(i, remainder)
            result.append(range(start_index, end_index))
            start_index = end_index

        assert len(result) == num_devices
        assert sum(len(x) for x in result) == total_size
        assert end_index == total_size

        return result

    def _tensor_preact(self, input_matrix, param_name, use_bias=True):
        """Helper function that creates a pre-activation of ``input_matrix`` by
        multiplying it by a weight matrix and adding a bias.

        ``input_matrix`` and the result normally have the shape of a mini-batch:
        the first dimension is the time step and the second dimension is the
        sequence. The last dimension is always the features. The preactivations
        for each mini-batch element will be computed by multiplying the features
        by the weight matrix and adding the bias. Thus the dimensionality of the
        input features should equal to the first dimension of the weight matrix,
        and the second dimension of the weight matrix defines the dimensionality
        of the output features.

        :type input_matrix: symbolic tensor
        :param input_matrix: one or more sequences of features in the shape
                             (time steps, sequences, features)

        :type param_name: str
        :param param_name: name of a parameter group that contains a weight
                           matrix and a bias vector

        :type use_bias: bool
        :param use_bias: if set to ``False``, does not add a bias

        :rtype: symbolic tensor
        :returns: a tensor that has the same number of dimensions as
                  ``input_matrix``, but the features (the last dimension) are
                  the preactivations
        """

        results = []
        for device in self._devices:
            weight = self._get_param(param_name + '/W', device)
            result = tensor.dot(input_matrix, weight)
            if use_bias:
                bias = self._get_param(param_name + '/b', device)
                result += bias
            results.append(result)

        if len(results) > 1:
            return tensor.concatenate(results, axis=2)
        elif len(results) == 1:
            return results[0]
        else:
            assert False

    def _tensor_conv1d(self, input_matrix, param_name):
        """Convolves ``input_matrix`` using filters and adds a bias.

        ``input_matrix`` and the result normally have the shape of a mini-batch:
        the first dimension is the time step and the second dimension is the
        sequence. The last dimension is always the features.

        The filter tensor is a stack of filters, each producing one output
        feature. One output feature in one mini-batch location will be computed
        by multiplying the input features in a local region by the corresponding
        filter elements, taking the sum over the input features in the region,
        and adding the bias. The first dimension of the filter defines the size
        of the region, the second dimension should equal to the dimensionality
        of the input features, and the third dimension defines the
        dimensionality of the output features.

        :type input_matrix: symbolic 3D tensor
        :param input_matrix: one or more sequences of features in the shape
                             (time steps, sequences, features)

        :type param_name: str
        :param param_name: name of a parameter group that contains a filter
                           matrix and a bias vector

        :rtype: symbolic 3D tensor
        :returns: the input convolved with the filters in the shape (time steps,
                  sequences, features)
        """

        # Permutate input dimensions from (time steps, sequences, features) to
        # (samples, elements, features).
        input_matrix = input_matrix.dimshuffle(1, 0, 2)

        results = []
        for device in self._devices:
            filters = self._get_param(param_name + '/W', device)
            result = conv1d(input_matrix,
                            filters,
                            padding='valid')
            results.append(result)

        if len(results) > 1:
            result = tensor.concatenate(results, axis=2)
        elif len(results) == 1:
            result = results[0]
        else:
            assert False

        # Permutate input dimensions from (samples, elements, features) to
        # (time steps, sequences, features).
        result = result.dimshuffle(1, 0, 2)

        result += self._get_param(param_name + '/b')
        return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements an iterator for reading mini-batches for scoring.
"""

import numpy

from theanolm.parsing.linearbatchiterator import LinearBatchIterator

class ScoringBatchIterator(LinearBatchIterator):
    """Iterator for Reading Mini-Batches for Scoring Text

    Returns the actual words in addition to the word IDs. These are needed for
    subword combination. File IDs are not returned.
    """

    def __init__(self, *args, **kwargs):
        """Constructs an iterator for reading mini-batches from given file or
        memory map.
        """

        super().__init__(*args, **kwargs)

    def _prepare_batch(self, sequences):
        """Transposes a list of sequences into a list of time steps. Then
        returns word ID and mask matrices in a format suitable to be input to
        the neural network, and words as a list of lists.

        The first returned matrix contains the word IDs. Word other than the
        shortlist words may be mapped to <unk>, depending on
        ``self._map_oos_to_unk``. The second matrix contains the words in plain
        text, and the third one contains a mask that defines which elements are
        past the sequence end. Where the other values are valid, the mask matrix
        contains ones.

        The word ID and mask matrices have the same shape. The first dimensions
        is the time step, i.e. the index to a word in a sequence. The second
        dimension selects the sequence. In other words, the first row is the
        first word of each sequence and so on. The plain text words are returned
        as a list of sequences, each sequence extending only to the sequence
        end.

        :type sequences: list of lists
        :param sequences: list of sequences, each of which is a list of word
                          IDs

        :rtype: three ndarrays
        :returns: word ID, word, and mask structures
        """

        num_sequences = len(sequences)
        batch_length = numpy.max([len(s) for s in sequences])

        unk_id = self._vocabulary.word_to_id['<unk>']
        shape = (batch_length, num_sequences)
        word_ids = numpy.ones(shape, numpy.int64) * unk_id
        words = []
        mask = numpy.zeros(shape, numpy.int8)

        for i, sequence in enumerate(sequences):
            length = len(sequence)

            sequence_word_ids = numpy.ones(length, numpy.int64) * unk_id
            sequence_words = []
            for index, (word, file_id) in enumerate(sequence):
                if word in self._vocabulary:
                    word_id = self._vocabulary.word_to_id[word]
                    if (not self._map_oos_to_unk) or \
                       self._vocabulary.in_shortlist(word_id):
                        sequence_word_ids[index] = word_id
                sequence_words.append(word)

            word_ids[:length, i] = sequence_word_ids
            words.append(sequence_words)
            mask[:length, i] = 1

        return word_ids, words, mask



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module for functions related to identifying the computation devices.
"""

import logging

import theano

def get_default_device(requested):
    """Enumerates the devices that have been configured for Theano and returns
    the name of the device that will be used by default.

    Devices and contexts are rather complicated in Theano.
    1) If no GPU is configured, the only device that can be used is cpu and any
       other requests will raise an exception.
    2) If a single GPU is enabled using ``device=cuda0``, the GPU device is
       ``None``, and only ``None`` or cpu can be selected.
    3) If multiple GPUs are enabled using ``contexts=dev0->cuda0;dev1->cuda1``
       and the default GPU is defined using ``device=cuda0``, then the default
       device will be ``None`` and dev0 does not exist. Either ``None``, dev1,
       or cpu can be selected.
    4) If multiple GPUs are enabled and the default GPU is not defined, then the
       ``None`` device does not exist. If ``None`` is requested, the first GPU
       device is returned.

    :type requested: str
    :param requested: ``None`` to automatically selected a default device; any
                      other selection either returns the same value or raises an
                      exception

    :rtype: str
    :returns: 'cpu' if no GPUs are enabled, otherwise ``None`` or the name of
              the first GPU device
    """

    try:
        names = []
        for name in theano.gpuarray.type.list_contexts():
            context = theano.gpuarray.type.get_context(name)
            logging.info('Context {} device="{}" ID="{}"'.format(
                         name, context.devname, context.unique_id))
            names.append(name)
        if requested is None:
            if any(name is None for name in names):
                return None
            elif len(names) > 0:
                logging.info('The first GPU device ("{}") will be used as the '
                             'default device.'.format(names[0]))
                return names[0]
            else:
                logging.info('Theano is not using a GPU or an old version of '
                             'libgpuarray is installed.')
                return None
        elif requested in names:
            logging.info('"{}" selected as the default device.'
                         .format(requested))
            return requested
        else:
            raise ValueError('Theano is not configured to use device "{}".'
                             .format(requested))
    except AttributeError:
        # No GPU support or we couldn't list the contexts because an old version
        # of libgpuarray is installed.
        if requested is None:
            logging.info('Theano is not using a GPU or an old version of '
                         'libgpuarray is installed.')
            return None
        elif requested == 'cpu':
            logging.info('Using CPU for computation.')
            return requested
        else:
            raise ValueError('Theano is not configured to use device "{}". '
                             'Only cpu is supported.'.format(requested))

def log_free_mem():
    """Writes the available GPU memory to the debug log.
    """

    for name in theano.gpuarray.type.list_contexts():
        context = theano.gpuarray.type.get_context(name)
        free_mbytes = context.free_gmem / (1024 * 1024)
        logging.debug("Available memory on GPU %s: %.0f MB", name, free_mbytes)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the Adam optimizer.
"""

import numpy
import theano
import theano.tensor as tensor

from theanolm.backend import Parameters
from theanolm.training.basicoptimizer import BasicOptimizer

class AdamOptimizer(BasicOptimizer):
    """Adam Optimization Method

    D. P. Kingma, J. Ba (2015)
    Adam: A Method for Stochastic Optimization
    The International Conference on Learning Representations (ICLR), San Diego
    """

    def __init__(self, optimization_options, network, *args, **kwargs):
        """Creates an Adam optimizer.

        :type optimization_options: dict
        :param optimization_options: a dictionary of optimization options

        :type network: Network
        :param network: the neural network object
        """

        self._params = Parameters()

        float_type = numpy.dtype(theano.config.floatX).type
        self._params.add('optimizer/timestep', float_type(0.0))

        for path, param in network.get_variables().items():
            self._params.add(path + '_mean_gradient',
                             numpy.zeros_like(param.get_value()))
            self._params.add(path + '_mean_sqr_gradient',
                             numpy.zeros_like(param.get_value()))

        # geometric rate for averaging gradients
        if 'gradient_decay_rate' not in optimization_options:
            raise ValueError("Gradient decay rate is not given in training "
                             "options.")
        self._gamma_m = optimization_options['gradient_decay_rate']

        # geometric rate for averaging squared gradients
        if 'sqr_gradient_decay_rate' not in optimization_options:
            raise ValueError("Squared gradient decay rate is not given in "
                             "optimization options.")
        self._gamma_ms = optimization_options['sqr_gradient_decay_rate']

        # momentum
        if 'momentum' not in optimization_options:
            raise ValueError("Momentum is not given in optimization options.")
        self._momentum = optimization_options['momentum']

        super().__init__(optimization_options, network, *args, **kwargs)

    def _get_param_updates(self, alpha):
        """Returns Theano expressions for updating the model parameters and any
        additional parameters required by the optimizer.

        :type alpha: Variable
        :param alpha: a scale to be applied to the model parameter updates

        :rtype: iterable over pairs (shared variable, new expression)
        :returns: expressions how to update the optimizer parameters
        """

        timestep_old = self._params['optimizer/timestep']
        timestep = timestep_old + 1.0
        alpha *= tensor.sqrt(1.0 - (self._gamma_ms ** timestep))
        alpha /= 1.0 - (self._gamma_m ** timestep)

        result = []
        deltas = dict()
        for path, gradient in zip(self.network.get_variables(),
                                  self._gradients):
            m_gradient_old = self._params[path + '_mean_gradient']
            ms_gradient_old = self._params[path + '_mean_sqr_gradient']
            m_gradient = \
                self._gamma_m * m_gradient_old + \
                (1.0 - self._gamma_m) * gradient
            ms_gradient = \
                self._gamma_ms * ms_gradient_old + \
                (1.0 - self._gamma_ms) * tensor.sqr(gradient)
            result.append((m_gradient_old, m_gradient))
            result.append((ms_gradient_old, ms_gradient))

            rms_gradient = tensor.sqrt(ms_gradient) + self._epsilon
            deltas[path] = -m_gradient / rms_gradient
        self._normalize(deltas)

        for path, param_old in self.network.get_variables().items():
            delta = deltas[path]
            result.append((param_old, param_old + alpha * delta))
        result.append((timestep_old, timestep))
        return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the Gated Linear Unit layer.
"""

import logging

import theano.tensor as tensor

from theanolm.network.basiclayer import BasicLayer
from theanolm.network.weightfunctions import get_submatrix

class GLULayer(BasicLayer):
    """Gated Linear Unit Layer

    Implements a convolution layer with a gating mechanism.

    Y. N. Dauphin (2017)
    Language Modeling with Gated Convolutional Networks
    Proc. International Conference on Machine Learning
    """

    def __init__(self, layer_options, *args, **kwargs):
        """Initializes the parameters used by this layer.
        """

        super().__init__(layer_options, *args, **kwargs)

        if 'filter_size' in layer_options:
            filter_size = int(layer_options['filter_size'])
        else:
            filter_size = 5
        logging.debug("  filter_size=%d", filter_size)

        assert len(self._input_layers) == 1
        input_size = self._input_layers[0].output_size
        output_size = self.output_size

        # concatenation of convolution filters for the linear projection and its
        # gate
        filter_shape = (filter_size, input_size, output_size)
        self._init_weight('input/W', filter_shape, scale=0.01, count=2)

        # concatenation of biases for the linear projection and its gate
        self._init_bias('input/b', output_size * 2)

        self._filter_size = filter_size
        self._input_size = input_size

        self.output = None

    def create_structure(self):
        """Creates the symbolic graph of this layer.

        The input is 3- or 4-dimensional: the first dimension is the time step,
        the second dimension are the sequences, and the third and fourth
        dimension are the layer input. The fourth dimension is created when more
        than one filter is used.
        """

        if not self._network.mode.minibatch:
            raise RuntimeError("Text generation and lattice decoding are not "
                               "possible with convolution layers.")

        layer_input = self._input_layers[0].output
        num_time_steps = layer_input.shape[0]
        num_sequences = layer_input.shape[1]
        input_size = self._input_size

        # Shift the input right by k - 1 time steps, where k is the filter size,
        # so that the output at any time step does not contain information from
        # future words.
        padding_size = self._filter_size - 1
        padding = tensor.zeros([padding_size, num_sequences, input_size])
        layer_input = tensor.concatenate([padding, layer_input])

        # Compute the linear projection and the gate pre-activation in a single
        # convolution operation.
        preact = self._tensor_conv1d(layer_input, 'input')
        linear = get_submatrix(preact, 0, self.output_size)
        gate = get_submatrix(preact, 1, self.output_size)

        self.output = linear * tensor.nnet.sigmoid(gate)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that defines file type classes to be used with command-line argument
parser.
"""

import io
import sys
import argparse
import gzip

class TextFileType(object):
    """An object that can be passed as the "type" argument to
    ArgumentParser.add_argument() in order to convert a path argument to a
    file object.

    If the path ends in ".gz", the file will be opened using gzip.open().
    UTF-8 encoding will be assumed. The special path "-" means standard
    input or output.

    Keyword Arguments:
      - mode -- A string indicating how the file is to be opened. Accepts the
        same values as the builtin open() function.
    """

    def __init__(self, mode='r'):
        self._mode = mode

    def __call__(self, string):
        if string is None:
            return None

        if string == '-':
            if 'r' in self._mode:
                return io.TextIOWrapper(sys.stdin.buffer, encoding='utf-8')
            elif ('w' in self._mode) or ('a' in self._mode):
                return io.TextIOWrapper(sys.stdout.buffer, encoding='utf-8')
            else:
                msg = "Cannot use standard input or output (file is opened " \
                      "with mode '%r')." % self._mode
                raise argparse.ArgumentTypeError(msg)

        try:
            if string.endswith('.gz'):
                return gzip.open(string, self._mode + 't', encoding='utf-8')
            return open(string, self._mode + 't', encoding='utf-8')
        except IOError as e:
            message = "Cannot open '%s': %s" % (string, e)
            raise argparse.ArgumentTypeError(message)
        except Exception as e:
            # I don't know a better way to communicate exceptions to the user.
            raise argparse.ArgumentTypeError(str(e))

    def __repr__(self):
        return '%s(%s)' % (type(self).__name__, self._mode)

class BinaryFileType(object):
    """An object that can be passed as the "type" argument to
    ArgumentParser.add_argument() in order to convert a path argument to a
    file object.

    If the path ends in ".gz", the file will be opened using gzip.open().
    UTF-8 encoding will be assumed. The special path "-" means standard
    input or output.

    Keyword Arguments:
      - mode -- A string indicating how the file is to be opened. Accepts the
        same values as the builtin open() function.
    """

    def __init__(self, mode='r'):
        self._mode = mode

    def __call__(self, string):
        if string is None:
            return None

        if string == '-':
            if 'r' in self._mode:
                return sys.stdin.buffer
            elif ('w' in self._mode) or ('a' in self._mode):
                return sys.stdout.buffer
            else:
                msg = "Cannot use standard input or output (file is opened " \
                      "with mode '%r')." % self._mode
                raise argparse.ArgumentTypeError(msg)

        try:
            if string.endswith('.gz'):
                return gzip.open(string, self._mode + 'b')
            return open(string, self._mode + 'b',)
        except IOError as e:
            message = "Cannot open '%s': %s" % (string, e)
            raise argparse.ArgumentTypeError(message)
        except Exception as e:
            # I don't know a better way to communicate exceptions to the user.
            raise argparse.ArgumentTypeError(str(e))

    def __repr__(self):
        return '%s(%s)' % (type(self).__name__, self._mode)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements classes for sampling noise words.
"""

from abc import abstractmethod, ABCMeta

import numpy
import theano
import theano.tensor as tensor


class ClassDistribution(object, metaclass=ABCMeta):
    """Base Class for Probability Distributions

    A probability distribution class implements methods for sampling words and
    converting words to probabilities.
    """

    def __init__(self, random):
        """Constructs a probability distribution.

        :type random: MRG_RandomStreams
        :param random: a random number generator
        """

        self._random = random
        self._float_type = numpy.dtype(theano.config.floatX).type

    @abstractmethod
    def sample(self, minibatch_size, num_samples):
        """Samples given number of class IDs per mini-batch element.
        """

        assert False

    @abstractmethod
    def probs(self, class_ids):
        """Converts class IDs to probabilities.

        The returned value may be a NumPy array or Theano tensor. It may be an
        array of the same shape as ``class_ids`` or a scalar which can be
        broadcasted to that shape.

        :type class_ids: numpy.ndarray
        :param class_ids: classes whose probabilities are requested

        :rtype: theano.config.floatX, numpy.ndarray, or Variable
        :returns: an array of the same shape as ``class_ids`` or a scalar which
                  can be broadcasted to that shape
        """

        assert False


class UniformDistribution(ClassDistribution):
    def __init__(self, random, support):
        """Constructs a uniform probability distribution.

        :type random: MRG_RandomStreams
        :param random: a random number generator

        :type support: int
        :param support: the sampled values will be in the range from 0 to
                        ``support - 1``
        """

        super().__init__(random)
        self._support = support

    def sample(self, minibatch_size, num_samples):
        """Samples given number of class IDs per mini-batch element.

        :type minibatch_size: int
        :param minibatch_size: number of mini-batch elements

        :type num_samples: int
        :param num_samples: number of samples to draw for each mini-batch element

        :rtype: Variable
        :returns: a 2-dimensional tensor variable describing the ``num_samples``
                  random samples for ``minibatch_size`` mini-batch element
        """

        sample = self._random.uniform(size=(minibatch_size, num_samples),
                                      high=self._support)
        return sample.astype('int64')

    def probs(self, class_ids):
        """Converts class IDs to probabilities.

        The returned value is a scalar. It will be broadcasted to the correct
        shape by Theano.

        :type class_ids: Variable
        :param class_ids: a symbolic vector describing the classes whose
                          probabilities are requested (ignored)

        :rtype: theano.config.floatX
        :returns: the probability of a single class
        """

        return self._float_type(1.0 / self._support)


class LogUniformDistribution(ClassDistribution):
    def __init__(self, random, support):
        """Constructs a probability distribution such that the logarithm of the
        samples is uniformly distributed.

        :type random: MRG_RandomStreams
        :param random: a random number generator

        :type support: int
        :param support: the sampled values will be in the range from 0 to
                        ``support - 1``
        """

        super().__init__(random)
        # Random numbers will be in the range [0, log(support + 1)[.
        self._log_support = numpy.log(support + 1)

    def sample(self, minibatch_size, num_samples):
        """Samples given number of class IDs per mini-batch element.

        :type minibatch_size: int
        :param minibatch_size: number of mini-batch elements

        :type num_samples: int
        :param num_samples: number of samples to draw for each mini-batch element

        :rtype: Variable
        :returns: a 2-dimensional tensor variable describing the ``num_samples``
                  random samples for ``minibatch_size`` mini-batch element
        """

        logs = self._random.uniform(size=(minibatch_size, num_samples),
                                    high=self._log_support)
        # The exponent will be in the range [1, support + 1[.
        sample = tensor.exp(logs) - 1
        return sample.astype('int64')

    def probs(self, class_ids):
        """Converts class IDs to probabilities.

        The returned value is a scalar. It will be broadcasted to the correct
        shape by Theano.

        :type class_ids: Variable
        :param class_ids: a symbolic vector describing the classes whose
                          probabilities are requested

        :rtype: Variable
        :returns: probabilities of the classes
        """

        # A sample will be in the range [x, x + 1[ when the log is in the range
        # [log(x + 1), log(x + 2)[. Thus the probability of x is
        # (log(x + 2) - log(x + 1)) / log_support.
        range = (class_ids + 2) / (class_ids + 1)
        return tensor.log(range) / self._log_support

class MultinomialDistribution(ClassDistribution):
    def __init__(self, random, probs):
        """Constructs a multinomial probability distribution.

        :type random: MRG_RandomStreams
        :param random: a random number generator

        :type probs: Variable
        :param probs: a tensor vector that defines the distribution where to
                      sample from
        """

        super().__init__(random)
        self._probs = probs

    def sample(self, minibatch_size, num_samples):
        """Samples given number of class IDs per mini-batch element.

        Theano supports currently only sampling without replacement. Thus if a
        different set of samples is required for each mini-batch element, we
        repeat the distribution for each mini-batch element.

        At some point the old interface, ``multinomial_wo_replacement()`` was
        faster, but is not supported anymore. In the future ``target='cpu'`` may
        improve the speed.

        :type minibatch_size: int
        :param minibatch_size: number of mini-batch elements

        :type num_samples: int
        :param num_samples: number of samples to draw for each mini-batch element

        :rtype: Variable
        :returns: a 2-dimensional tensor variable describing the ``num_samples``
                  random samples for ``minibatch_size`` mini-batch element
        """

        probs = self._probs[None, :]
        probs = tensor.tile(probs, [minibatch_size, 1])
#        return self._random.multinomial_wo_replacement(pvals=probs, n=num_samples)
        sample = self._random.choice(size=num_samples, replace=False, p=probs,
                                     target='cpu')
        # Some versions of Theano seem to return crazy high or low numbers because
        # of some rounding errors, so we take the modulo to be safe.
        sample %= probs.shape[1]
        return sample

    def probs(self, class_ids):
        """Converts class IDs to probabilities.

        The returned value is a scalar. It will be broadcasted to the correct
        shape by Theano.

        :type class_ids: Variable
        :param class_ids: a symbolic vector describing the classes whose
                          probabilities are requested

        :rtype: Variable
        :returns: symbolic vector describing the probabilities of the classes
        """

        return self._probs[class_ids]



# -*- coding: utf-8 -*-
"""Executable that launches wctool command.
"""

from traceback import format_tb
import logging
import sys

from wordclasses import main

def exception_handler(exception_type, exception, traceback):
    """Exception handler. Writes stack trace in case of uncaught exceptions to
    debug log.
    """

    print("{}: {}".format(exception_type.__name__, exception))
    logging.debug("Traceback:")
    for item in format_tb(traceback):
        logging.debug(item.strip())
    sys.exit(2)

if __name__ == "__main__":
    sys.excepthook = exception_handler
    try:
        main()
    except KeyboardInterrupt:
        print('Interrupted from keyboard.')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import numpy
from wordclasses.functions import byte_size

class BigramOptimizer(object):
    """Word Class Optimizer
    """

    def __init__(self, vocabulary, count_type = 'int32'):
        """Reads the vocabulary from the text ``corpus_file`` and creates word
        IDs. The vocabulary may be restricted by ``vocabulary_file``.

        :type corpus_file: file object
        :param corpus_file: a file that contains the input sentences

        :type vocabulary: theanolm.Vocabulary
        :param vocabulary: words to include in the optimization and initial classes
        """

        self._initial_vocabulary = vocabulary
        self.vocabulary_size = vocabulary.num_shortlist_words()
        self.num_classes = vocabulary.num_classes()
        self.num_normal_classes = vocabulary.num_normal_classes
        self._count_type = count_type

    def move_to_best_class(self, word):
        """Moves a word to the class that minimizes training set log likelihood.

        :type word: int
        :param word: ID of the word to be moved
        """

        if word.startswith('<') and word.endswith('>'):
            return False

        word_id = self.get_word_id(word)
        old_class_id = self.get_word_class(word_id)
        if self._class_size(old_class_id) < 2:
            logging.debug('Less than two words in class %d. Not moving word %s.',
                          old_class_id, word)
            return False

        ll_diff, new_class_id = self._find_best_move(word_id)
        if ll_diff > 0:
            self._move(word_id, new_class_id)
            return True
        else:
            return False

    def _compute_class_statistics(self, word_counts, ww_counts, word_to_class):
        """Computes class statistics from word statistics given the
        word-to-class mapping.

        :type word_counts: numpy.ndarray
        :param word_counts: word unigram counts

        :type ww_counts: scipy.sparse.csc_matrix
        :param ww_counts: word bigram counts

        :type word_to_class: numpy.ndarray
        :param word_to_class: gives the class ID of each word ID
        """

        class_counts = numpy.zeros(self.num_classes, dtype=self._count_type)
        logging.debug("Allocated %s for class counts.",
                      byte_size(class_counts.nbytes))
        cc_counts = numpy.zeros(
            (self.num_classes, self.num_classes), dtype=self._count_type)
        logging.debug("Allocated %s for class-class counts.",
                      byte_size(cc_counts.nbytes))
        cw_counts = numpy.zeros(
            (self.num_classes, self.vocabulary_size), dtype=self._count_type)
        logging.debug("Allocated %s for class-word counts.",
                      byte_size(cw_counts.nbytes))
        wc_counts = numpy.zeros(
            (self.vocabulary_size, self.num_classes), dtype=self._count_type)
        logging.debug("Allocated %s for word-class counts.",
                      byte_size(wc_counts.nbytes))

        numpy.add.at(class_counts, word_to_class, word_counts)

        left_word_ids, right_word_ids = ww_counts.nonzero()
        counts = ww_counts[left_word_ids, right_word_ids].flat
        left_class_ids = word_to_class[left_word_ids]
        right_class_ids = word_to_class[right_word_ids]
        numpy.add.at(cc_counts, (left_class_ids, right_class_ids), counts)
        numpy.add.at(cw_counts, (left_class_ids, right_word_ids), counts)
        numpy.add.at(wc_counts, (left_word_ids, right_class_ids), counts)

        return class_counts, cc_counts, cw_counts, wc_counts

    def _find_best_move(self, word_id):
        """Finds the class such that moving the given word to that class would
        give best improvement in log likelihood.

        :type word_id: int
        :param word_id: ID of the word to be moved

        :rtype: (float, int)
        :returns: a tuple containing the amount log likelihood would change and
                  the ID of the class where the word should be moved (or None if
                  there was just one word in the class)
        """

        best_ll_diff = -numpy.inf
        best_class_id = None

        old_class_id = self.get_word_class(word_id)
        for class_id in range(self.num_normal_classes):
            if class_id == old_class_id:
                continue
            ll_diff = self._evaluate(word_id, class_id)
            if ll_diff > best_ll_diff:
                best_ll_diff = ll_diff
                best_class_id = class_id

        return best_ll_diff, best_class_id

    def _evaluate(self, word_id, new_class_id):
        """Evaluates how much moving a word to another class would change the
        log likelihood.

        :type word_id: int
        :param word_id: ID of the word to be moved

        :type new_class_id: int
        :param new_class_id: ID of the class the word will be moved to

        :rtype: int
        :returns: log likelihood change
        """

        raise NotImplementedError("BigramOptimizer._evaluate() has to be "
                                  "implemented by the subclass.")

    def _move(self, word_id, new_class_id):
        """Moves a word to another class.

        :type word_id: int
        :param word_id: ID of the word to be moved

        :type new_class_id: int
        :param new_class_id: ID of the class the word will be moved to
        """

        raise NotImplementedError("BigramOptimizer._move() has to be "
                                  "implemented by the subclass.")

    def get_word_id(self, word):
        """Returns the ID of the given word.

        :type word: str
        :param word: a word

        :rtype: int
        :returns: ID of the given word
        """

        return self._initial_vocabulary.word_to_id[word]

    def words(self):
        """A generator for iterating through the words.

        :rtype: generator for (str, int, float)
        :returns: generates a tuple containing a word, its class ID, and unigram
                  class membership probability
        """

        for word, word_id in self._initial_vocabulary.word_to_id.items():
            class_id = self.get_word_class(word_id)
            prob = self.get_word_prob(word_id)
            yield word, class_id, prob

    def get_word_class(self, word_id):
        """Returns the class the given word is currently assigned to.

        :type word_id: int
        :param word_id: ID of the word

        :rtype: int
        :returns: ID of the class the word is assigned to
        """

        raise NotImplementedError("BigramOptimizer.get_word_class() has to be "
                                  "implemented by the subclass.")

    def get_word_prob(self, word_id):
        """Returns the unigram probability of a word within its class.

        :type word_id: int
        :param word_id: ID of the word

        :rtype: float
        :returns: class membership probability
        """

        raise NotImplementedError("BigramOptimizer.get_word_prob() has to be "
                                  "implemented by the subclass.")

    def log_likelihood(self):
        """Computes the log likelihood that a bigram model would give to the
        corpus.

        :rtype: float
        :returns: log likelihood of the training corpus
        """

        raise NotImplementedError("BigramOptimizer.log_likelihood() has to be "
                                  "implemented by the subclass.")

    def _class_size(self, class_id):
        """Calculates the number of words in a class.

        :type class_id: int
        :param class_id: ID of a class

        :rtype: int
        :returns: number of words in the class
        """

        raise NotImplementedError("BigramOptimizer._class_size() has to be "
                                  "implemented by the subclass.")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Modules related to vocabulary, word classes, and word statistics.
"""

from theanolm.vocabulary.vocabulary import Vocabulary
from theanolm.vocabulary.statistics import compute_word_counts, BigramStatistics



print("import: 'theanolm'")
import theanolm

print("import: 'theanolm.commands'")
import theanolm.commands

print("import: 'theanolm.network'")
import theanolm.network

print("import: 'theanolm.parsing'")
import theanolm.parsing

print("import: 'theanolm.scoring'")
import theanolm.scoring

print("import: 'theanolm.training'")
import theanolm.training

print("import: 'theanolm.vocabulary'")
import theanolm.vocabulary




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the TextSampler class.
"""

import numpy
import theano
from theanolm.network import RecurrentState

class TextSampler(object):
    """Neural network language model sampler

    A Theano function that generates text using a neural network language
    model.
    """

    def __init__(self, network):
        """Creates a Theano function that samples the next word of a set of word
        sequences.

        Creates the function self.step_function that takes as input a set of
        word sequences and the current recurrent states. It uses the previous
        states and word IDs to compute the output distributions. It samples from
        the output distributions and returns the sampled word IDs along with the
        output states of this time step.

        :type network: Network
        :param network: the neural network object
        """

        self._network = network
        self._vocabulary = network.vocabulary
        self._random = network.random

        inputs = [network.input_word_ids, network.input_class_ids]
        inputs.extend(network.recurrent_state_input)

        # multinomial() is only implemented for dimension <= 2, but the matrix
        # contains only one time step anyway.
        output_probs = network.output_probs()[0]
        # Theano doesn't set a test value to some of it's internal variables.
        # Disable the warning since there's nothing we can do.
        compute_test_value = theano.config.compute_test_value
        theano.config.compute_test_value = 'off'
        class_ids = self._random.multinomial(pvals=output_probs).argmax(1)
        class_ids.tag.test_value = numpy.zeros(4, dtype='int64')
        theano.config.compute_test_value = compute_test_value
        class_ids = class_ids.reshape([1, class_ids.shape[0]])
        outputs = [class_ids]
        outputs.extend(network.recurrent_state_output)

        # Ignore unused input, because is_training is only used by dropout
        # layer.
        self.step_function = theano.function(
            inputs,
            outputs,
            givens=[(network.is_training, numpy.int8(0))],
            name='step_sampler',
            on_unused_input='ignore')

    def generate(self, length, num_sequences=1, seed_sequence=''):
        """Generates a text sequence.

        Calls self.step_function() repeatedly, reading the word output and
        the state output of the hidden layer and passing the hidden layer state
        output to the next time step.

        :type length: int
        :param length: number of words (tokens) in each sequence

        :type num_sequences: int
        :param num_sequences: number of sequences to generate in parallel

        :rtype: list of list of strs
        :returns: list of word sequences
        """
        seed_tokens = seed_sequence.strip().split()
        sos_id = self._vocabulary.word_to_id['<s>']
        sos_class_id = self._vocabulary.word_id_to_class_id[sos_id]

        input_word_ids = sos_id * \
                         numpy.ones(shape=(1, num_sequences)).astype('int64')
        input_class_ids = sos_class_id * \
                          numpy.ones(shape=(1, num_sequences)).astype('int64')
        result = sos_id * \
                 numpy.ones(shape=(len(seed_tokens)+length,
                   num_sequences)).astype('int64')
        state = RecurrentState(self._network.recurrent_state_size,
                               num_sequences)
       
        #First, possibly compute forward passes with the seed sequence
        for time_step, token in enumerate(seed_tokens, start=1):
            step_result = self.step_function(input_word_ids,
                                             input_class_ids,
                                             *state.get())
            token_id = self._vocabulary.word_to_id[token]
            token_class_id = \
                self._vocabulary.word_id_to_class_id[token_id]
            input_word_ids = token_id * \
                             numpy.ones(shape=(1,      
                               num_sequences)).astype('int64')
            input_class_ids = token_class_id * \
                              numpy.ones(shape=(1,
                                num_sequences)).astype('int64')
            step_word_ids = input_word_ids
            result[time_step] = step_word_ids
            state.set(step_result[1:])
        
        #Then sample:
        for time_step in range(len(seed_tokens) + 1, length+len(seed_tokens)):
            # the input is the output from the previous step.
            step_result = self.step_function(input_word_ids,
                                             input_class_ids,
                                             *state.get())
            class_ids = step_result[0]
            # The class IDs from the single time step.
            step_class_ids = class_ids[0]
            step_word_ids = numpy.array(
                self._vocabulary.class_ids_to_word_ids(step_class_ids))
            result[time_step] = step_word_ids
            input_word_ids = step_word_ids[numpy.newaxis]
            input_class_ids = class_ids
            state.set(step_result[1:])

        return self._vocabulary.id_to_word[result.transpose()].tolist()
